"""
    An example script to digitize/reconstruct/clustering endcap ecal hits
"""
from Gaudi.Configuration import *
import json
import os
import ROOT

from Configurables import ApplicationMgr, EICDataSvc, PodioInput, PodioOutput, GeoSvc
from GaudiKernel.SystemOfUnits import MeV, GeV, mm, cm, mrad

detector_name = str(os.environ.get("DETECTOR", "epic"))
detector_config = str(os.environ.get("DETECTOR_CONFIG", detector_name))
detector_version = str(os.environ.get("DETECTOR_VERSION", "main"))
detector_path = str(os.environ.get("DETECTOR_PATH", "."))
compact_path = str(
    os.environ.get(
        "JUGGLER_COMPACT_PATH",
        "{}.xml".format(os.path.join(detector_path, detector_name)),
    )
)

# Detector features that affect reconstruction
has_ecal_barrel_scfi = False
if "epic" in detector_name and "imaging" in detector_config:
    has_ecal_barrel_scfi = True
if 'epic' in detector_name and 'brycecanyon' in detector_config:
    has_ecal_barrel_scfi = True

# input and output
input_sims = [
    f.strip() for f in str.split(os.environ["JUGGLER_SIM_FILE"], ",") if f.strip()
]
output_rec = str(os.environ["JUGGLER_REC_FILE"])
n_events = int(os.environ["JUGGLER_N_EVENTS"])

# input arguments from calibration file
with open(f"{detector_path}/calibrations/emcal_barrel_calibration.json") as f:
    calib_data = json.load(f)["electron"]

cb_ecal_sf = float(calib_data["sampling_fraction_img"])

# geometry service
geo_service = GeoSvc("GeoSvc", detectors=[compact_path], OutputLevel=INFO)
# data service
podioevent = EICDataSvc("EventDataSvc", inputs=input_sims)


# juggler components
from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsMerger as CalHitsMerger
from Configurables import Jug__Reco__CalorimeterIslandCluster as IslandCluster
from Configurables import Jug__Reco__ImagingPixelReco as ImCalPixelReco
from Configurables import Jug__Reco__ImagingTopoCluster as ImagingCluster
from Configurables import Jug__Reco__ImagingClusterReco as ImagingClusterReco
from Configurables import Jug__Reco__ClusterRecoCoG as RecoCoG

# branches needed from simulation root file
sim_coll = [
    "MCParticles",
]

if has_ecal_barrel_scfi:
    sim_coll += [
        "EcalBarrelImagingHits",
        "EcalBarrelImagingHitsContributions",
    ]
else:
    sim_coll += [
        "EcalBarrelSciGlassHits",
        "EcalBarrelSciGlassHitsContributions",
    ]

algorithms = []

# input and output
podin = PodioInput("PodioReader", collections=sim_coll)
algorithms.append(podin)
podout = PodioOutput("out", filename=output_rec)
algorithms.append(podout)

# Central Barrel Ecal
if has_ecal_barrel_scfi:
    # Central Barrel Ecal (Imaging Cal.)
    cb_ecal_daq = dict(
        dynamicRangeADC=3 * MeV, capacityADC=8192, pedestalMean=400, pedestalSigma=20
    )  # about 6 keV

    cb_ecal_digi = CalHitDigi(
        "cb_ecal_digi",
        inputHitCollection="EcalBarrelImagingHits",
        outputHitCollection="EcalBarrelImagingHitsDigi",
        energyResolutions=[0.0, 0.02, 0.0],  # 2% flat resolution
        **cb_ecal_daq,
    )
    algorithms.append(cb_ecal_digi)

    cb_ecal_reco = ImCalPixelReco(
        "cb_ecal_reco",
        inputHitCollection=cb_ecal_digi.outputHitCollection,
        outputHitCollection="EcalBarrelImagingHitsReco",
        samplingFraction=cb_ecal_sf,
        thresholdFactor=3,  # about 20 keV
        readoutClass="EcalBarrelImagingHits",  # readout class
        layerField="layer",  # field to get layer id
        sectorField="module",  # field to get sector id
        **cb_ecal_daq,
    )
    algorithms.append(cb_ecal_reco)

    cb_ecal_cl = ImagingCluster(
        "cb_ecal_cl",
        inputHitCollection=cb_ecal_reco.outputHitCollection,
        outputProtoClusterCollection="EcalBarrelImagingProtoClusters",
        localDistXY=[2.0 * mm, 2 * mm],  # same layer
        layerDistEtaPhi=[10 * mrad, 10 * mrad],  # adjacent layer
        neighbourLayersRange=2,  # id diff for adjacent layer
        sectorDist=3.0 * cm,
    )  # different sector
    algorithms.append(cb_ecal_cl)

    cb_ecal_clreco = RecoCoG(
        "cb_ecal_clreco",
        inputProtoClusterCollection=cb_ecal_cl.outputProtoClusterCollection,
        outputClusterCollection="EcalBarrelImagingClusters",
    )
    algorithms.append(cb_ecal_clreco)

else:
    # SciGlass calorimeter
    sciglass_ecal_daq = dict(
        dynamicRangeADC=5.0 * GeV, capacityADC=32768, pedestalMean=400, pedestalSigma=3
    )

    sciglass_ecal_digi = CalHitDigi(
        "sciglass_ecal_digi",
        inputHitCollection="EcalBarrelSciGlassHits",
        outputHitCollection="EcalBarrelSciGlassHitsDigi",
        energyResolutions=[0.0, 0.02, 0.0],  # 2% flat resolution
        **sciglass_ecal_daq,
    )
    algorithms.append(sciglass_ecal_digi)

    sciglass_ecal_reco = CalHitReco(
        "sciglass_ecal_reco",
        inputHitCollection=sciglass_ecal_digi.outputHitCollection,
        outputHitCollection="EcalBarrelSciGlassHitsReco",
        thresholdFactor=3,  # about 20 keV
        readoutClass="EcalBarrelSciGlassHits",  # readout class
        sectorField="sector",  # field to get sector id
        samplingFraction=0.998,  # this accounts for a small fraction of leakage
        **sciglass_ecal_daq,
    )
    algorithms.append(sciglass_ecal_reco)

    sciglass_ecal_cl = IslandCluster(
        "sciglass_ecal_cl",
        inputHitCollection=sciglass_ecal_reco.outputHitCollection,
        outputProtoClusterCollection="EcalBarrelProtoClusters",
        splitCluster=False,
        minClusterHitEdep=1.0 * MeV,  # discard low energy hits
        minClusterCenterEdep=30 * MeV,
        # Magic constants:
        #  24 - number of sectors
        #  5  - number of towers per sector
        adjacencyMatrix = "(abs(tower_1 - tower_2) + (abs((sector_1 - sector_2) * 5 + row_1 - row_2) == 1) + (abs((sector_1 - sector_2) * 5 + row_1 - row_2) == (24 * 5 - 1))) == 1",
        readoutClass = "EcalBarrelSciGlassHits",
    )
    algorithms.append(sciglass_ecal_cl)

    sciglass_ecal_clreco = RecoCoG(
        "sciglass_ecal_clreco",
        inputProtoClusterCollection=sciglass_ecal_cl.outputProtoClusterCollection,
        # mcHits="EcalBarrelSciGlassHits",
        outputClusterCollection="EcalBarrelClusters",
        #OutputLevel=DEBUG
    )
    algorithms.append(sciglass_ecal_clreco)

    # sciglass_barrel_clreco = RecoCoG(
    #    "sciglass_barrel_clreco",
    #    inputProtoClusterCollection=sciglass_ecal_cl.outputProtoClusterCollection,
    #    outputClusterCollection="EcalBarrelScFiClusters",
    #    logWeightBase=6.2,
    # )
    # algorithms.append(sciglass_barrel_clreco)


podout.outputCommands = [
    "drop *",
    "keep MCParticles",
    "keep *HitsReco",
    "keep *HitsDigi",
    "keep *Cluster*",
]

ApplicationMgr(
    TopAlg=algorithms,
    EvtSel="NONE",
    EvtMax=n_events,
    ExtSvc=[podioevent],
    OutputLevel=DEBUG,
)
