"""
    An example script to digitize/reconstruct/clustering endcap ecal hits
"""
from Gaudi.Configuration import *
import os
import ROOT

from Configurables import ApplicationMgr, EICDataSvc, PodioInput, PodioOutput, GeoSvc
from GaudiKernel.SystemOfUnits import MeV, GeV, mm, cm, mrad

detector_name = str(os.environ.get("DETECTOR_CONFIG", "epic"))
detector_path = str(os.environ.get("DETECTOR_PATH", "."))
compact_path = str(
    os.environ.get(
        "JUGGLER_COMPACT_PATH",
        "{}.xml".format(os.path.join(detector_path, detector_name)),
    )
)
ci_ecal_sf = float(os.environ.get("CI_ECAL_SAMP_FRAC", 0.253))

# input and output
input_sims = [
    f.strip() for f in str.split(os.environ["JUGGLER_SIM_FILE"], ",") if f.strip()
]
output_rec = str(os.environ["JUGGLER_REC_FILE"])
n_events = int(os.environ["JUGGLER_N_EVENTS"])

# geometry service
geo_service = GeoSvc("GeoSvc", detectors=[compact_path], OutputLevel=INFO)
# data service
podioevent = EICDataSvc("EventDataSvc", inputs=input_sims)


# juggler components
from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsMerger as CalHitsMerger
from Configurables import Jug__Reco__CalorimeterIslandCluster as IslandCluster
from Configurables import Jug__Reco__ClusterRecoCoG as RecoCoG

# branches needed from simulation root file
sim_coll = [
    "MCParticles",
    "EcalEndcapPHits",
    "EcalEndcapPHitsContributions",
]

# input and output
podin = PodioInput("PodioReader", collections=sim_coll)
podout = PodioOutput("out", filename=output_rec)


# Endcap sampling Ecal
ci_ecal_daq = dict(
    dynamicRangeADC=50.0 * MeV, capacityADC=32768, pedestalMean=400, pedestalSigma=10
)

ci_ecal_digi = CalHitDigi(
    "ci_ecal_digi",
    inputHitCollection="EcalEndcapPHits",
    outputHitCollection="EcalEndcapPHitsDigi",
    **ci_ecal_daq
)

ci_ecal_reco = CalHitReco(
    "ci_ecal_reco",
    inputHitCollection=ci_ecal_digi.outputHitCollection,
    outputHitCollection="EcalEndcapPHitsReco",
    samplingFraction=ci_ecal_sf,
    thresholdFactor=5.0,
    **ci_ecal_daq
)

# merge hits in different layer (projection to local x-y plane)
ci_ecal_merger = CalHitsMerger(
    "ci_ecal_merger",
    # OutputLevel=DEBUG,
    inputHitCollection=ci_ecal_reco.outputHitCollection,
    outputHitCollection="EcalEndcapPHitsRecoXY",
    fields=["fiber_x", "fiber_y"],
    fieldRefNumbers=[1, 1],
    # fields=["layer", "slice"],
    # fieldRefNumbers=[1, 0],
    readoutClass="EcalEndcapPHits",
)

ci_ecal_cl = IslandCluster(
    "ci_ecal_cl",
    # OutputLevel=DEBUG,
    inputHitCollection=ci_ecal_merger.outputHitCollection,
    outputProtoClusterCollection="EcalEndcapPProtoClusters",
    splitCluster=False,
    minClusterCenterEdep=10.0 * MeV,
    localDistXY=[10 * mm, 10 * mm],
)

ci_ecal_clreco = RecoCoG(
    "ci_ecal_clreco",
    inputProtoClusterCollection=ci_ecal_cl.outputProtoClusterCollection,
    outputClusterCollection="EcalEndcapPClusters",
    logWeightBase=6.2,
)

podout.outputCommands = [
    "drop *",
    "keep MCParticles",
    "keep *HitsReco*",
    "keep *HitsDigi",
    "keep *Cluster*",
]

ApplicationMgr(
    TopAlg=[
        podin,
        ci_ecal_digi,
        ci_ecal_reco,
        ci_ecal_merger,
        ci_ecal_cl,
        ci_ecal_clreco,
        podout,
    ],
    EvtSel="NONE",
    EvtMax=n_events,
    ExtSvc=[podioevent],
    OutputLevel=DEBUG,
)
