import os
import ROOT
from Gaudi.Configuration import *
from GaudiKernel.SystemOfUnits import mm, MeV, rad, ns
from Configurables import ApplicationMgr, EICDataSvc, PodioOutput, GeoSvc

from Configurables import PodioInput
from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsMerger as CalHitsMerge
from Configurables import Jug__Reco__CalorimeterIslandCluster as IslandCluster
from Configurables import Jug__Reco__ClusterRecoCoG as RecoCoG

kwargs = dict()

# input arguments through environment variables
kwargs['input'] = os.environ.get('JUGGLER_SIM_FILE', '../epic/zdc_photons.edm4hep.root')
kwargs['output'] = os.environ.get('JUGGLER_REC_FILE', 'rec_zdc_photons.root')
kwargs['compact'] = os.environ.get('DETECTOR_COMPACT_PATH', '../epic/epic.xml')
kwargs['nev'] = int(os.environ.get('JUGGLER_N_EVENTS', 100))
kwargs['PbSci_sf'] = float(os.environ.get('ZDC_PbSCI_SAMP_FRAC', 1.0))

if kwargs['nev'] < 1:
    f = ROOT.TFile(kwargs['input'])
    kwargs['nev'] = f.events.GetEntries()

print(kwargs)

geo_service = GeoSvc('GeoSvc', detectors=kwargs['compact'].split(','), OutputLevel=INFO)
podioevent = EICDataSvc('EventDataSvc', inputs=kwargs['input'].split(','), OutputLevel=DEBUG)

sim_colls = [
    'MCParticles',
    'ZDCEcalHits',
    'ZDCEcalHitsContributions',
    'ZDCHcalHits',
    'ZDCHcalHitsContributions',
]

podin = PodioInput('PodioReader', collections=sim_colls, OutputLevel=DEBUG)
podout = PodioOutput('out', filename=kwargs['output'])

# Digitization
WSciFi_digi = CalHitDigi('WSciFi_digi',
        inputHitCollection='ZDCEcalHits',
        outputHitCollection='ZDCEcalRawHits')

PbSci_digi = CalHitDigi('PbSci_digi',
        inputHitCollection='ZDCHcalHits',
        outputHitCollection='ZDCHcalRawHits')

# Reconstruction
WSciFi_reco = CalHitReco('WSciFi_reco',
        inputHitCollection=WSciFi_digi.outputHitCollection,
        outputHitCollection='ZDCEcalRecHits',
        readoutClass='ZDCEcalHits',
        localDetFields=['system'])

PbSci_reco = CalHitReco('PbSci_reco',
        inputHitCollection=PbSci_digi.outputHitCollection,
        outputHitCollection='ZDCHcalRecHits',
        readoutClass='ZDCHcalHits',
        localDetFields=['system'])

# Clustering
WSciFi_cl = IslandCluster('WSciFi_cl',
        inputHitCollection=WSciFi_reco.outputHitCollection,
        outputProtoClusterCollection='ZDCEcalProtoClusters',
        minClusterCenterEdep=3.*MeV,
        minClusterHitEdep=0.1*MeV,
        localDistXY=[50*mm, 50*mm],
        splitCluster=True)

WSciFi_clreco = RecoCoG('WSciFi_clreco',
        inputProtoClusterCollection = WSciFi_cl.outputProtoClusterCollection,
        outputClusterCollection='ZDCEcalClusters',
        logWeightBase=3.6)

PbSci_cl = IslandCluster('PbSci_cl',
        inputHitCollection=PbSci_reco.outputHitCollection,
        outputProtoClusterCollection='ZDCHcalProtoClusters',
        minClusterCenterEdep=1.*MeV,
        minClusterHitEdep=0.1*MeV,
        localDistXY=[200*mm, 200*mm],
        splitCluster=False)

PbSci_clreco = RecoCoG('PbSci_clreco',
        inputProtoClusterCollection = PbSci_cl.outputProtoClusterCollection,
        outputClusterCollection='ZDCHcalClusters',
        logWeightBase=3.6,
        samplingFraction=kwargs['PbSci_sf'])

podout.outputCommands = ['keep *']

ApplicationMgr(
    TopAlg=[podin,
        WSciFi_digi, PbSci_digi, 
        WSciFi_reco, PbSci_reco,
        WSciFi_cl, WSciFi_clreco, PbSci_cl, PbSci_clreco,
        podout],
    EvtSel='NONE',
    EvtMax=kwargs['nev'],
    ExtSvc=[podioevent],
    OutputLevel=DEBUG
)

