#!/bin/bash

function print_the_help {
  echo "USAGE: ${0} [--rec-only] [--ana-only] [--particle=neutron] "
  echo "  OPTIONS: "
  echo "            --rec-only   only run gaudi reconstruction part"
  echo "            --ana-only   only run analysis macros part"
  exit 
}

PARTICLE="neutron"
REC_ONLY=
ANALYSIS_ONLY=

POSITIONAL=()
while [[ $# -gt 0 ]]
do
  key="$1"

  case $key in
    -h|--help)
      shift # past argument
      print_the_help
      ;;
    --rec-only)
      REC_ONLY=1
      shift # past value
      ;;
    --ana-only)
      ANALYSIS_ONLY=1
      shift # past value
      ;;
    --particle)
      PARTICLE="$2"
      shift # past argument
      shift # past value
      ;;
    *)    # unknown option
      #POSITIONAL+=("$1") # save it in an array for later
      echo "unknown option $1"
      print_the_help
      shift # past argument
      ;;
  esac
done
set -- "${POSITIONAL[@]}" # restore positional parameters


print_env.sh

## To run the reconstruction, we need the following global variables:
## - JUGGLER_INSTALL_PREFIX:   Install prefix for Juggler (simu/recon)
## - DETECTOR:         the detector package we want to use for this benchmark
## - DETECTOR_VERSION: the detector package we want to use for this benchmark
## - DETECTOR_PATH:            full path to the detector definitions
##
## You can ready options/env.sh for more in-depth explanations of the variables
## and how they can be controlled.
export DETECTOR_PATH=${DETECTOR_PATH}

if [[ ! -n  "${JUGGLER_N_EVENTS}" ]] ; then 
  export JUGGLER_N_EVENTS=100
fi

if [[ ! -n  "${E_start}" ]] ; then
  export E_start=10.0
fi

if [[ ! -n  "${E_end}" ]] ; then
  export E_end=10.0
fi

if [[ ! -n "${ZDC_SCI_SAMP_FRAC}" ]] ; then
  export ZDC_PbSCI_SAMP_FRAC=1.0
fi

export DETECTOR_COMPACT_PATH=${DETECTOR_PATH}/${DETECTOR_CONFIG}.xml

export JUGGLER_FILE_NAME_TAG="zdc_${PARTICLE}"
export JUGGLER_GEN_FILE="${JUGGLER_FILE_NAME_TAG}.hepmc"

export JUGGLER_SIM_FILE="sim_${JUGGLER_FILE_NAME_TAG}.edm4hep.root"
export JUGGLER_REC_FILE="rec_${JUGGLER_FILE_NAME_TAG}.root"

export RESULTS_PATH="results/far_forward/zdc/${PARTICLE}"
mkdir -p ${RESULTS_PATH}

echo "JUGGLER_N_EVENTS = ${JUGGLER_N_EVENTS}"
echo "DETECTOR = ${DETECTOR}"

if [[ -z "${REC_ONLY}" && -z "${ANALYSIS_ONLY}" ]] ;
then
  echo "Generating input events"
  root -b -q "benchmarks/far_forward/analysis/gen_zdc_particles.cxx(${JUGGLER_N_EVENTS}, \"${PARTICLE}\", ${E_start}, ${E_end}, \"${JUGGLER_FILE_NAME_TAG}.hepmc\")"
  if [[ "$?" -ne "0" ]] ; then
    echo "ERROR running script: generating input events"
    exit 1
  fi

  echo "Running Geant4 simulation"
  ddsim --runType batch \
    --part.minimalKineticEnergy 0.5*MeV  \
    --filter.tracker edep0 \
    -v WARNING \
    --numberOfEvents ${JUGGLER_N_EVENTS} \
    --compactFile ${DETECTOR_PATH}/${DETECTOR_CONFIG}.xml \
    --inputFiles  ${JUGGLER_FILE_NAME_TAG}.hepmc \
    --outputFile  ${JUGGLER_SIM_FILE}
  if [[ "$?" -ne "0" ]] ; then
    echo "ERROR running npdet: Geant4 simulation"
    exit 1
  fi
fi

rootls -t ${JUGGLER_SIM_FILE}

if [[ -z "${ANALYSIS_ONLY}" ]] ;
then
  gaudirun.py benchmarks/far_forward/options/zdc_reconstruction.py
  if [[ "$?" -ne "0" ]] ; then
    echo "ERROR running juggler"
    exit 1
  fi
fi

rootls -t ${JUGGLER_REC_FILE}

echo "Running analysis root scripts"
root -b -q "benchmarks/far_forward/analysis/analysis_zdc_particles.cxx+(\"${JUGGLER_REC_FILE}\", \"${RESULTS_PATH}\")"
if [[ "$?" -ne "0" ]] ; then
  echo "ERROR running root analysis script"
  exit 1
fi

root_filesize=$(stat --format=%s "${JUGGLER_REC_FILE}")
if [[ "${JUGGLER_N_EVENTS}" -lt "500" ]] ; then
  # file must be less than 10 MB to upload
  if [[ "${root_filesize}" -lt "10000000" ]] ; then
    cp ${JUGGLER_SIM_FILE} results/far_forward/zdc/.
    cp ${JUGGLER_REC_FILE} results/far_forward/zdc/.
  fi
fi
