import json
import os
import ROOT
from Gaudi.Configuration import *
from GaudiKernel.SystemOfUnits import GeV, MeV, mm, cm, mrad, rad, ns
from Configurables import ApplicationMgr, EICDataSvc, PodioOutput, GeoSvc

from Configurables import PodioInput
from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsMerger as CalHitsMerger
from Configurables import Jug__Reco__CalorimeterIslandCluster as IslandCluster
from Configurables import Jug__Reco__ClusterRecoCoG as RecoCoG
from Configurables import Jug__Reco__ImagingPixelReco as ImagingPixelReco
from Configurables import Jug__Reco__ImagingTopoCluster as ImagingTopoCluster
from Configurables import Jug__Reco__ImagingClusterReco as ImagingClusterReco

detector_name = str(os.environ.get("DETECTOR", "epic"))
detector_config = str(os.environ.get("DETECTOR_CONFIG", detector_name))
detector_version = str(os.environ.get("DETECTOR_VERSION", "main"))

detector_path = str(os.environ.get("DETECTOR_PATH", "."))

# Detector features that affect reconstruction
has_ecal_barrel_scfi = False
if 'epic' in detector_name and 'imaging' in detector_config:
    has_ecal_barrel_scfi = True
if 'epic' in detector_name and 'brycecanyon' in detector_config:
    has_ecal_barrel_scfi = True

# input arguments from calibration file
with open(f'{detector_path}/calibrations/emcal_barrel_calibration.json') as f:
    calib_data = json.load(f)['electron']

#print(calib_data)

kwargs = dict()
kwargs['img_sf'] = float(calib_data['sampling_fraction_img'])
kwargs['scfi_sf'] = float(calib_data['sampling_fraction_scfi'])
#kwargs['img_sf'] = 0.0060795
#kwargs['scfi_sf'] = 0.10343

# input arguments through environment variables
kwargs['input'] = os.environ['CB_EMCAL_SIM_FILE']
kwargs['output'] = os.environ['CB_EMCAL_REC_FILE']
kwargs['compact'] = os.environ['CB_EMCAL_COMPACT_PATH']
kwargs['nev'] = int(os.environ['CB_EMCAL_NUMEV'])

if kwargs['nev'] < 1:
    f = ROOT.TFile(kwargs['input'])
    kwargs['nev'] = f.events.GetEntries()

print(kwargs)

# get sampling fraction from system environment variable, 1.0 by default
geo_service = GeoSvc('GeoSvc', detectors=kwargs['compact'].split(','), OutputLevel=INFO)
podioevent = EICDataSvc('EventDataSvc', inputs=kwargs['input'].split(','), OutputLevel=DEBUG)

sim_colls = [
    'MCParticles',
]
ecal_barrel_scfi_collections = [
    'EcalBarrelImagingHits',
    'EcalBarrelImagingHitsContributions',
    'EcalBarrelScFiHits',
    'EcalBarrelScFiHitsContributions'
]
ecal_barrel_sciglass_collections = [
    'EcalBarrelSciGlassHits',
    'EcalBarrelSciGlassHitsContributions',
]
if has_ecal_barrel_scfi:
    sim_colls += ecal_barrel_scfi_collections
else:
    sim_colls += ecal_barrel_sciglass_collections

# list of algorithms
algorithms = []

podin = PodioInput('PodioReader', collections=sim_colls, OutputLevel=DEBUG)
algorithms.append(podin)


# Central Barrel Ecal
if has_ecal_barrel_scfi:
    # Central ECAL Imaging Calorimeter

    # use the same daq_setting for digi/reco pair
    imcaldaq = dict(
        dynamicRangeADC=3*MeV,
        capacityADC=32767,
        pedestalMean=400,
        pedestalSigma=50)   # 50/32767*3 MeV ~ 5 keV

    imcaldigi = CalHitDigi('imcal_digi',
        OutputLevel=DEBUG,
        inputHitCollection='EcalBarrelImagingHits',
        outputHitCollection='DigiEcalBarrelImagingHits',
        energyResolutions=[0., 0.02, 0.],
        **imcaldaq)
    algorithms.append(imcaldigi)

    imcalreco = ImagingPixelReco('imcal_reco',
        #OutputLevel=DEBUG,
        inputHitCollection=imcaldigi.outputHitCollection,
        outputHitCollection='RecoEcalBarrelImagingHits',
        readoutClass='EcalBarrelImagingHits',
        layerField='layer',
        sectorField='module',
        samplingFraction=kwargs['img_sf'],
        **imcaldaq)
    algorithms.append(imcalreco)

    imcalcluster = ImagingTopoCluster('imcal_cluster',
        #OutputLevel=DEBUG,
        inputHitCollection=imcalreco.outputHitCollection,
        outputProtoClusterCollection='EcalBarrelImagingProtoClusters',
        localDistXY=[2.*mm, 2*mm],
        layerDistEtaPhi=[10*mrad, 10*mrad],
        neighbourLayersRange=2,
        minClusterEdep= 0.5*MeV/kwargs['img_sf'],
        sectorDist=3.*cm)
    algorithms.append(imcalcluster)

    clusterreco = ImagingClusterReco('imcal_clreco',
        #OutputLevel=DEBUG,
        inputProtoClusters=imcalcluster.outputProtoClusterCollection,
        outputLayers='EcalBarrelImagingClustersLayers',
        outputClusters='EcalBarrelImagingClusters',
        mcHits="EcalBarrelImagingHits")
    algorithms.append(clusterreco)

    # scfi layers
    scfi_barrel_daq = dict(
        dynamicRangeADC=50.*MeV,
        capacityADC=32768,
        pedestalMean=400,
        pedestalSigma=10)

    scfi_barrel_digi = CalHitDigi("scfi_barrel_digi",
        inputHitCollection="EcalBarrelScFiHits",
        outputHitCollection="EcalBarrelScFiHitsDigi",
        **scfi_barrel_daq)
    algorithms.append(scfi_barrel_digi)

    scfi_barrel_reco = CalHitReco("scfi_barrel_reco",
        inputHitCollection=scfi_barrel_digi.outputHitCollection,
        outputHitCollection="EcalBarrelScFiHitsReco",
        thresholdFactor=5.0,
        readoutClass="EcalBarrelScFiHits",
        layerField="layer",
        sectorField="module",
        localDetFields=["system", "module"], # use local coordinates in each module (stave)
        samplingFraction=kwargs['scfi_sf'],
        **scfi_barrel_daq)
    algorithms.append(scfi_barrel_reco)

    # merge hits in different layer (projection to local x-y plane)
    scfi_barrel_merger = CalHitsMerger("scfi_barrel_merger",
        #OutputLevel=DEBUG,
        inputHitCollection=scfi_barrel_reco.outputHitCollection,
        outputHitCollection="EcalBarrelScFiGridReco",
        fields=["fiber", "z"],
        fieldRefNumbers=[1, 1],
        readoutClass="EcalBarrelScFiHits")
    algorithms.append(scfi_barrel_merger)

    scfi_barrel_cl = IslandCluster("scfi_barrel_cl",
        #OutputLevel=DEBUG,
        inputHitCollection=scfi_barrel_merger.outputHitCollection,
        outputProtoClusterCollection="EcalBarrelScFiProtoClusters",
        splitCluster=False,
        minClusterCenterEdep=10.*MeV,
        localDistXZ=[30*mm, 30*mm],
        sectorDist=5.*cm)
    algorithms.append(scfi_barrel_cl)

    scfi_barrel_clreco = RecoCoG("scfi_barrel_clreco",
        inputProtoClusterCollection=scfi_barrel_cl.outputProtoClusterCollection,
        outputClusterCollection="EcalBarrelScFiClusters",
        logWeightBase=6.2)
    algorithms.append(scfi_barrel_clreco)

else:
    # SciGlass calorimeter
    sciglass_ecal_daq = dict(
        dynamicRangeADC=5.*GeV,
        capacityADC=32768,
        pedestalMean=400,
        pedestalSigma=3)

    sciglass_ecal_digi = CalHitDigi("sciglass_ecal_digi",
        inputHitCollection="EcalBarrelSciGlassHits",
        outputHitCollection="EcalBarrelSciGlassHitsDigi",
        energyResolutions=[0., 0.02, 0.],   # 2% flat resolution
        **sciglass_ecal_daq)
    algorithms.append(sciglass_ecal_digi)

    sciglass_ecal_reco = CalHitReco("sciglass_ecal_reco",
        inputHitCollection=sciglass_ecal_digi.outputHitCollection,
        outputHitCollection="EcalBarrelSciGlassHitsReco",
        thresholdFactor=3,  # about 20 keV
        readoutClass="EcalBarrelSciGlassHits",  # readout class
        sectorField="sector",           # field to get sector id
        samplingFraction=0.998,         # this accounts for a small fraction of leakage
        **sciglass_ecal_daq)
    algorithms.append(sciglass_ecal_reco)

    sciglass_ecal_cl = IslandCluster("sciglass_ecal_cl",
        inputHitCollection=sciglass_ecal_reco.outputHitCollection,
        outputProtoClusterCollection="EcalBarrelProtoClusters",
        splitCluster=False,
        minClusterHitEdep=1.0*MeV,  # discard low energy hits
        minClusterCenterEdep=30*MeV,
        # Magic constants:
        #  24 - number of sectors
        #  5  - number of towers per sector
        adjacencyMatrix = "(abs(tower_1 - tower_2) + (abs((sector_1 - sector_2) * 5 + row_1 - row_2) == 1) + (abs((sector_1 - sector_2) * 5 + row_1 - row_2) == (24 * 5 - 1))) == 1",
        readoutClass = "EcalBarrelSciGlassHits")
    algorithms.append(sciglass_ecal_cl)

    sciglass_ecal_clreco = ImagingClusterReco("sciglass_ecal_clreco",
        inputProtoClusters=sciglass_ecal_cl.outputProtoClusterCollection,
        mcHits="EcalBarrelSciGlassHits",
        outputClusters="EcalBarrelClusters",
        outputLayers="EcalBarrelLayers")
    algorithms.append(sciglass_ecal_clreco)


# TODO: merge two types of clusters

podout = PodioOutput('podout', filename=kwargs['output'])
podout.outputCommands = ['drop *',
        'keep MCParticles',
        'keep *Reco*',
        'keep *Digi*',
        'keep *Cluster*',
        'keep *Layer*',
        ]
algorithms.append(podout)

ApplicationMgr(
    TopAlg=algorithms,
    EvtSel='NONE',
    EvtMax=kwargs['nev'],
    ExtSvc=[podioevent],
    OutputLevel=DEBUG
)
