import json
import os
import ROOT
from Gaudi.Configuration import *
from GaudiKernel.SystemOfUnits import mm, MeV, rad, ns
from Configurables import ApplicationMgr, EICDataSvc, PodioOutput, GeoSvc

from Configurables import PodioInput
from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsEtaPhiProjector as CalHitsProj
from Configurables import Jug__Reco__CalorimeterIslandCluster as IslandCluster
from Configurables import Jug__Reco__ClusterRecoCoG as RecoCoG

detector_path = str(os.environ.get("DETECTOR_PATH", "."))

# input arguments from calibration file
with open(f'{detector_path}/calibrations/emcal_barrel_calibration.json') as f:
    calib_data = json.load(f)['electron']

kwargs = dict()
kwargs['img_sf'] = float(calib_data['sampling_fraction_img'])

# input arguments through environment variables
kwargs['input'] = os.environ.get('CB_EMCAL_SIM_FILE', 'barrel_pion0_5GeV.edm4hep.root')
kwargs['output'] = os.environ.get('CB_EMCAL_REC_FILE', 'barrel_pion0_5GeV_cluster.root')
kwargs['compact'] = os.environ.get('CB_EMCAL_COMPACT_PATH', 'epic/epic.xml')
kwargs['nev'] = int(os.environ.get('CB_EMCAL_NUMEV', 100))

if kwargs['nev'] < 1:
    f = ROOT.TFile(kwargs['input'])
    kwargs['nev'] = f.events.GetEntries()

print(kwargs)

geo_service = GeoSvc('GeoSvc', detectors=kwargs['compact'].split(','), OutputLevel=INFO)
podioevent = EICDataSvc('EventDataSvc', inputs=kwargs['input'].split(','), OutputLevel=DEBUG)

podin = PodioInput('PodioReader', collections=['MCParticles', 'EcalBarrelImagingHits',  'EcalBarrelImagingHitsContributions'], OutputLevel=DEBUG)
podout = PodioOutput('out', filename=kwargs['output'])


# use the same daq_setting for digi/reco pair
imcal_barrel_daq = dict(
        dynamicRangeADC=3.*MeV,
        capacityADC=32768,
        pedestalMean=400,
        pedestalSigma=50)

imcal_barrel_digi = CalHitDigi('imcal_barrel_digi',
        inputHitCollection='EcalBarrelImagingHits',
        outputHitCollection='EcalBarrelImagingHitsDigi',
        **imcal_barrel_daq)

imcal_barrel_reco = CalHitReco('imcal_barrel_reco',
        inputHitCollection=imcal_barrel_digi.outputHitCollection,
        outputHitCollection='EcalBarrelImagingHitsReco',
        thresholdFactor=5.0,
        readoutClass='EcalBarrelImagingHits',
        localDetFields=['system', 'module'], # use local coordinates in each module (stave)
        **imcal_barrel_daq)

# merge hits to eta, phi bins
imcal_barrel_merger = CalHitsProj('imcal_barrel_merger',
        # OutputLevel=DEBUG,
        inputHitCollection=imcal_barrel_reco.outputHitCollection,
        outputHitCollection='EcalBarrelImagingHitsMerge',
        gridSizes=[0.004, 0.004*rad])

imcal_barrel_cl = IslandCluster('imcal_barrel_cl',
        # OutputLevel=DEBUG,
        inputHitCollection=imcal_barrel_merger.outputHitCollection,
        outputProtoClusterCollection='EcalBarrelImagingProtoClusters',
        minClusterCenterEdep=1.*MeV,
        minClusterHitEdep=0.1*MeV,
        splitCluster=True,
        globalDistEtaPhi=[0.005, 0.005*rad])

imcal_barrel_clreco = RecoCoG('imcal_barrel_clreco',
        # OutputLevel=DEBUG,
        inputProtoClusterCollection = imcal_barrel_cl.outputProtoClusterCollection,
        outputClusterCollection='EcalBarrelImagingClusters',
        logWeightBase=6.2,
        samplingFraction=kwargs['img_sf'])


podout.outputCommands = ['keep *']

ApplicationMgr(
    TopAlg=[podin,
        imcal_barrel_digi, imcal_barrel_reco,
        imcal_barrel_merger, imcal_barrel_cl, imcal_barrel_clreco,
        podout],
    EvtSel='NONE',
    EvtMax=kwargs['nev'],
    ExtSvc=[podioevent],
    OutputLevel=DEBUG
)


