'''
    A simple analysis script to check raw hits vs. merged hits
    *  ReadoutDecoder class solely depends on python xml parser and is thus much faster, but it may not work
       if some of the features are not supported by the parser (but supported by dd4hep parser)
    ** dd4hep decoder can also be used (more reliable but much slower), check get_layer_ids.py for the code
    Chao Peng (ANL)
'''
import os
import ROOT
import pandas as pd
import numpy as np
import argparse
from matplotlib import pyplot as plt
import matplotlib.ticker as ticker
from utils import ReadoutDecoder, flatten_collection


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('rec_file', help='Path to reconstruction output file.')
    parser.add_argument('-o', dest='outdir', default='.', help='Output directory.')
    parser.add_argument('-c', '--compact', dest='compact', required=True,
            help='Top-level xml file of the detector description')
    parser.add_argument('--raw-coll', dest='rcoll', required=True,
            help='Raw hits collection name in the reconstruction file')
    parser.add_argument('--merged-coll', dest='mcoll', required=True,
            help='Merged hits collection name in the reconstruction file')
    parser.add_argument('--readout', dest='readout', required=True,
            help='Readout name for the hits collection')
    parser.add_argument('--merged-fields', dest='mfields', required=True,
            help='Merged fields (separated by \",\")')
    args = parser.parse_args()

    # decoder and merge mask
    decoder = ReadoutDecoder(args.compact, args.readout)
    mmask = np.uint64(0)
    for field in args.mfields.split(','):
        fmask = decoder.mask(field.strip())
        mmask = np.bitwise_or(mmask, fmask)
        print('{:20s} {:#066b}'.format('Field Mask - ' + field, fmask))
    mmask = np.bitwise_not(mmask)
    print('{:20s} {:#066b}'.format('Merging Mask', mmask))

    # build dataframes from reconstructed root files
    # get hits
    columns = ['cellID', 'position.x', 'position.y', 'position.z', 'energy']
    rdf_rec = ROOT.RDataFrame('events', args.rec_file)
    # raw hits
    dfr = flatten_collection(rdf_rec.Range(100), args.rcoll, columns)
    dfr.rename(columns={c: c.replace(args.rcoll + '.', '') for c in dfr.columns}, inplace=True)
    # merged hits
    dfm = flatten_collection(rdf_rec, args.mcoll, columns)
    dfm.rename(columns={c: c.replace(args.mcoll + '.', '') for c in dfm.columns}, inplace=True)

    dfr.loc[:, 'masked_id'] = np.bitwise_and(mmask, dfr['cellID'])
    dfm.loc[:, 'merged_id'] = np.bitwise_and(mmask, dfm['cellID'])

    dfr = dfr.merge(dfm.groupby('event')['merged_id'].apply(np.array), on='event')
    dfr.loc[:, 'raw_merge_num'] = dfr.apply(lambda r: np.count_nonzero(r.masked_id == r.merged_id), axis=1)

    dfm = dfm.merge(dfr.groupby('event')['masked_id'].apply(np.array), on='event')
    dfm.loc[:, 'merge_raw_num'] = dfm.apply(lambda r: np.count_nonzero(r.merged_id == r.masked_id), axis=1)

    print(dfr[['masked_id', 'merged_id', 'raw_merge_num']])
    print(dfm[['masked_id', 'merged_id', 'merge_raw_num']])

    # dfr.to_csv('raw_hits.csv')
    # dfm.to_csv('merged_hits.csv')

    # merged hits (grid)
    # dfm = pd.read_csv('merged_hits.csv', index_col=0)
    ev_grpm = dfm.groupby('event')
    # energy scaling (assume mean at 5 GeV)
    escale = 5000./np.mean(ev_grpm['energy'].sum())
    dfm['energy'] = dfm['energy']*escale
    ev_ngrids = ev_grpm['energy'].size()
    nev = len(ev_ngrids)

    # number of grids per event
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(ev_ngrids, ec='k', bins=30, rwidth=1.0, weights=np.repeat(1./nev, nev))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('# of Grids per Event', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('grid_per_event.png')

    # energy deposit per grid
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(dfm['energy'], ec='k', bins=30, rwidth=1.0, weights=np.repeat(1./len(dfm), len(dfm)))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('Grid $E_{dep}$ (MeV)', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('edep_per_grid.png')

    # number of fired fibers per grid
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(dfm['merge_raw_num'], ec='k', bins=30, rwidth=1.0, weights=np.repeat(1./len(dfm), len(dfm)))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('# of Fiber-Signals per Grid', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('signals_per_grid.png')


    # raw hits (fiber)
    # dfr = pd.read_csv('raw_hits.csv', index_col=0)
    dfr['energy'] = dfr['energy']*escale
    ev_grpr = dfr.groupby('event')
    ev_nfibers = ev_grpr['energy'].size()
    nev = len(ev_nfibers)

    # number of fiber per event
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(ev_nfibers, ec='k', bins=30, rwidth=1.0, weights=np.repeat(1./nev, nev))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('# of Fibers per Event', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('fiber_per_event.png')

    # energy deposit per fiber
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(dfr['energy'], ec='k', bins=30, rwidth=1.0, weights=np.repeat(1./len(dfr), len(dfr)))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('Fiber $E_{dep}$ (MeV)', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('edep_per_fiber.png')

    # number of fired fibers per grid
    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.hist(dfr['raw_merge_num'], ec='k', bins=np.arange(-0.5, 7.5, 1.0),
            rwidth=1.0, weights=np.repeat(1./len(dfr), len(dfr)))
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$', fontsize=26)
    ax.set_xlabel('# of Grids per Fiber-Signal', fontsize=24)
    ax.set_ylabel('Normalized Counts', fontsize=24)
    fig.savefig('grids_per_fiber.png')

    # build relations between two datasets
    dfe = dfr.groupby(['event', 'masked_id'])['energy'].sum().reset_index()
    print(dfe)
    dfe.rename(columns={'energy':'energy_fibers'}, inplace=True)
    dfem = dfm[['event', 'merged_id', 'energy']].rename(columns={'merged_id': 'masked_id'})
    dfe = dfe.merge(dfem, on=['event', 'masked_id'])

    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    ax.scatter(*dfe[['energy_fibers', 'energy']].values.T)
    ax.tick_params(labelsize=24)
    ax.set_title('5.0 GeV $e^-$ (100 events)', fontsize=26)
    ax.set_xlabel('Fiber Group Energy Sum (MeV)', fontsize=24)
    ax.set_ylabel('Grid Energy (MeV)', fontsize=24)
    fig.savefig('fiber_grid_energy.png')


