import ROOT
import os
import gc
import numpy as np
import pandas as pd
import argparse
import sys


# read from RDataFrame and flatten a given collection, return pandas dataframe
def flatten_collection(rdf, collection, cols=None, event_colname='event'):
    if not cols:
        cols = [str(c) for c in rdf.GetColumnNames() if str(c).startswith('{}.'.format(collection))]
    else:
        cols = ['{}.{}'.format(collection, c) for c in cols]
    if not cols:
        print('cannot find any branch under collection {}'.format(collection))
        return pd.DataFrame()

    data = rdf.AsNumpy(cols)
    # flatten the data, add an event id to identify clusters from different events
    evns = []
    for i, vec in enumerate(data[cols[0]]):
        evns += [i]*vec.size()
    for n, vals in data.items():
        # make sure ints are not converted to floats
        typename = vals[0].__class__.__name__.lower()
        dtype = np.int64 if 'int' in typename or 'long' in typename else np.float32
        # type safe creation
        data[n] = np.asarray([v for vec in vals for v in vec], dtype=dtype)
    # build data frame
    dfp = pd.DataFrame({c: pd.Series(v) for c, v in data.items()})
    dfp.loc[:, event_colname] = evns
    return dfp


# load root macros, input is an argument string
def load_root_macros(arg_macros):
    for path in arg_macros.split(','):
        path = path.strip()
        if os.path.exists(path):
            ROOT.gROOT.Macro(path)
        else:
            print('\"{}\" does not exist, skip loading it.'.format(path))


def cartesian_to_polar(x, y, z):
    r = np.sqrt(x**2 + y**2 + z**2)
    rc = np.sqrt(x**2 + y**2)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)
    eta = -np.log(np.tan(theta / 2.))
    return r, theta, phi, rc, eta


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Visualize the cluster (layer-wise) from analysis')
    parser.add_argument('file', type=str, help='path to root file')
    parser.add_argument('-o', type=str, default='.', dest='outdir', help='output directory')
    # parser.add_argument('--compact', type=str, default='', dest='compact', help='compact file')
    parser.add_argument('-m', '--macros', type=str, default='rootlogon.C', dest='macros',
                         help='root macros to load (accept multiple paths separated by \",\")')
    parser.add_argument('--topo-range', type=float, default=1000.0, dest='topo_range',
                        help='half range for projection plot (mrad)')
    parser.add_argument('--shape', type=str, default='20, 20', dest='shape',
                        help='shape of data (Nlayer, Nhits), default (20, 20)')
    parser.add_argument('--features', type=str, default='energy, eta, phi',
                        help='features of hit data, default (eta, phi, energy)')
    parser.add_argument('--branch', type=str, default='EcalBarrelImagingHitsML', help='name of data branch')
    parser.add_argument('--null-columns', type=str, default='', help='make the columns zeros')
    args = parser.parse_args()

    os.makedirs(args.outdir, exist_ok=True)
    load_root_macros(args.macros)

    # read data and MCParticles
    rdf = ROOT.RDataFrame("events", args.file)
    df = flatten_collection(rdf, args.branch, ['layer', 'energy', 'position.x', 'position.y', 'position.z'])
    df.rename(columns={c: c.replace(args.branch + '.', '') for c in df.columns}, inplace=True)

    r, theta, phi, rc, eta = cartesian_to_polar(*df[['position.x', 'position.y', 'position.z']].values.T)
    df.loc[:, 'r'] = r
    df.loc[:, 'theta'] = theta
    df.loc[:, 'phi'] = phi
    df.loc[:, 'rc'] = rc
    df.loc[:, 'eta'] = eta

    dfm = flatten_collection(rdf, 'MCParticles', ['generatorStatus', 'PDG', 'momentum.x', 'momentum.y', 'momentum.z', 'mass'])
    dfm.rename(columns={c: c.replace('MCParticles.', '') for c in dfm.columns}, inplace=True)
    # selete incident particles
    dfm = dfm[dfm['generatorStatus'].isin([0, 1])]
    # NOTE: assumed single particles
    dfm = dfm.groupby('event').first()
    p, theta, phi, pT, eta = cartesian_to_polar(*dfm[['momentum.x', 'momentum.y', 'momentum.z']].values.T)
    dfm.loc[:, 'p'] = p
    dfm.loc[:, 'theta'] = theta
    dfm.loc[:, 'phi'] = phi
    dfm.loc[:, 'pT'] = pT
    dfm.loc[:, 'eta'] = eta

    # energy unit is in GeV, converting to MeV first
    df.loc[:, 'eMeV'] = df['energy']*1000
    # scaled by 5 MeV to have most of them in [0, 1]
    df.loc[:, 'energy'] = df['eMeV']/5.
    # convert eta, phi, edep so they are within [0, 1]
    df.loc[:, 'etotal'] = df.groupby('event')['energy'].transform('sum')
    df = df[df['etotal'] > 0.]

    event_group = df.groupby('event')
    df.loc[:, 'logw'] = 0.
    # energy = 0 means no hits
    mask = df['energy'] > 0.
    df.loc[mask, 'logw'] = np.log(df.loc[mask, 'energy'].values / df.loc[mask, 'etotal'].values) + 6.2
    df.loc[:, 'logw'] = df['logw'].clip(0)
    df.loc[:, 'wtotal'] = event_group['logw'].transform('sum')

    df.loc[:, 'etaw'] = df['eta']*df['logw']
    df.loc[:, 'phiw'] = df['phi']*df['logw']

    df.loc[:, 'eta_mean'] = event_group['etaw'].transform('sum') / df['wtotal']
    df.loc[:, 'phi_mean'] = event_group['phiw'].transform('sum') / df['wtotal']

    # topo range in rad
    trg = args.topo_range / 1000.
    df.loc[mask, 'eta'] = (df.loc[mask, 'eta'].values - df.loc[mask, 'eta_mean'].values + trg) / (2.*trg)
    df.loc[mask, 'phi'] = ((df.loc[mask, 'phi'].values - df.loc[mask, 'phi_mean'].values)/np.pi + trg) / (2.*trg)
    df.loc[~mask, 'eta'] = 0.
    df.loc[~mask, 'phi'] = 0.
    df.loc[~mask, 'rc'] = 0.
    # scale by 2 meter, which keeps barrel region well under 1.0
    df.loc[:, 'rc'] = df['rc'].values / 2000.0
    df.loc[:, 'layer_type'] = np.floor(df['layer']/100.).astype(int)
    df.loc[df['layer_type'] == 1, 'eta'] = 0.

    # null specified columns (for scfi which does not have eta info)
    nullcols = [c.strip() for c in args.null_columns.split(',')]
    for col in nullcols:
        df.loc[:, col] = 0.

    gc.collect()
    # features and data-shape
    featcols = [f.strip() for f in args.features.split(',')]
    dshape = [int(i.strip()) for i in args.shape.split(',')]
    dshape.append(len(featcols))

    event_ids = df['event'].unique()
    data = df.set_index('event')[featcols].clip(0, 1).values.reshape([len(event_ids)] + dshape)

    tags = dfm.loc[event_ids, ['PDG', 'p', 'pT', 'eta', 'phi', 'mass']]
    # also save Edep per layers
    # merge the sandwich layers
    df['layer'] = df['layer'] % 100
    edeps = df.groupby(['event', 'layer'])['eMeV'].sum().unstack()

    """ TEST
    from matplotlib import pyplot as plt

    fig, axs = plt.subplots(3, 1)
    axs.flat[0].hist(df.loc[df['edep'] > 0, 'eta'].values, bins=np.linspace(-2, 2, 101), ec='k')
    axs.flat[1].hist(df.loc[df['edep'] > 0, 'phi'].values, bins=np.linspace(-2, 2, 101), ec='k')
    axs.flat[2].hist(df.loc[df['edep'] > 0, 'edep'].values, bins=np.linspace(-2, 2, 101), ec='k')
    fig.savefig('test.png')
    """
    print('Training data set shape = {}'.format(data.shape))
    print('Tagging data set shape = {}'.format(tags.shape))
    print('Edep/layer data set shape = {}'.format(edeps.shape))

    np.save(os.path.join(args.outdir, 'data.npy'), data)
    np.save(os.path.join(args.outdir, 'tag.npy'), tags.values)
    np.save(os.path.join(args.outdir, 'edep.npy'), edeps.values)

