import os
import ROOT
from Gaudi.Configuration import *
from GaudiKernel import SystemOfUnits as units

from Configurables import ApplicationMgr, EICDataSvc, PodioInput, PodioOutput, GeoSvc
from GaudiKernel.SystemOfUnits import MeV, GeV, mm, cm, mrad

from Configurables import Jug__Digi__CalorimeterHitDigi as CalHitDigi
from Configurables import Jug__Reco__CalorimeterHitReco as CalHitReco
from Configurables import Jug__Reco__CalorimeterHitsMerger as CalHitsMerger
# from Configurables import Jug__Reco__ImagingPixelMerger as MLDataMerger
# from Configurables import Jug__Reco__ImagingPixelDataSorter as MLDataSorter
# from Configurables import Jug__Reco__ImagingPixelDataCombiner as MLDataCombiner


# input arguments through environment variables
kwargs = dict()
kwargs['input'] = os.environ.get('JUGGLER_SIM_FILE', '')
kwargs['output'] = os.environ.get('JUGGLER_REC_FILE', '')
kwargs['compact'] = os.environ.get('JUGGLER_COMPACT_PATH', '')
kwargs['nev'] = int(os.environ.get('JUGGLER_N_EVENTS', -1))

if kwargs['nev'] < 1:
    f = ROOT.TFile(kwargs['input'])
    kwargs['nev'] = f.events.GetEntries()
print(kwargs)
# get sampling fraction from system environment variable, 1.0 by default
sf = float(os.environ.get('JUGGLER_SAMP_FRAC', '1.0'))

geo_service  = GeoSvc('GeoSvc', detectors=[f.strip() for f in kwargs['compact'].split(',')])
podev = EICDataSvc('EventDataSvc', inputs=[f.strip() for f in kwargs['input'].split(',')])

podin = PodioInput('PodioReader', collections=[
    'MCParticles',
    'EcalBarrelImagingHits',
    'EcalBarrelImagingHitsContributions',
    'EcalBarrelScFiHits',
    'EcalBarrelScFiHitsContributions'
    ])
podout = PodioOutput('out', filename=kwargs['output'])

# Barrel Ecal AstroPix
becal_img_daq = dict(
        dynamicRangeADC=3*MeV,
        capacityADC=2**13,
        pedestalMean=400,
        pedestalSigma=20
        )   # about 6 keV pedestal

becal_img_digi = CalHitDigi(
        'becal_img_digi',
        inputHitCollection='EcalBarrelImagingHits',
        outputHitCollection='EcalBarrelImagingHitsDigi',
        energyResolutions=[0., 0.02, 0.],   # 2% flat resolution
        **becal_img_daq
        )

becal_img_reco = CalHitReco(
        'becal_img_reco',
        inputHitCollection=becal_img_digi.outputHitCollection,
        outputHitCollection='EcalBarrelImagingHitsReco',
        thresholdFactor=3,  # about 20 keV
        readoutClass='EcalBarrelImagingHits',  # readout class
        layerField='layer',             # field to get layer id
        sectorField='module',           # field to get sector id
        **becal_img_daq
        )

"""
becal_img_merger = MLDataMerger(
        'becal_img_merger',
        inputHits=becal_img_reco.outputHitCollection,
        outputHits='EcalBarrelImagingHitsSeg',
        etaSize=0.001,
        phiSize=0.001
        )

becal_img_sorter = MLDataSorter('becal_img_sorter',
        inputHitCollection=becal_img_merger.outputHits,
        outputHitCollection='EcalBarrelImagingHitsML',
        numberOfLayers=6,
        numberOfHits=20
        )
"""

# Barrel ECAL ScFi
# use the same daq_setting for digi/reco pair
becal_scfi_daq = dict(
        dynamicRangeADC=50.*MeV,
        capacityADC=2**15,
        pedestalMean=400,
        pedestalSigma=10)

becal_scfi_digi = CalHitDigi('becal_scfi_digi',
        inputHitCollection='EcalBarrelScFiHits',
        outputHitCollection='EcalBarrelScFiHitsDigi',
        **becal_scfi_daq)

becal_scfi_reco = CalHitReco('becal_scfi_reco',
        inputHitCollection=becal_scfi_digi.outputHitCollection,
        outputHitCollection='EcalBarrelScFiHitsReco',
        thresholdFactor=5.0,
        readoutClass='EcalBarrelScFiHits',
        layerField='layer',
        sectorField='module',
        localDetFields=['system', 'module'], # use local coordinates in each module (stave)
        **becal_scfi_daq)

# merge hits in different layer (projection to local x-y plane)
becal_scfi_merger = CalHitsMerger('becal_scfi_merger',
        # OutputLevel=DEBUG,
        inputHitCollection=becal_scfi_reco.outputHitCollection,
        outputHitCollection='EcalBarrelScFiGridReco',
        fields=['fiber', 'z'],
        fieldRefNumbers=[1, 1],
        readoutClass='EcalBarrelScFiHits')

"""
becal_scfi_sorter = MLDataSorter('becal_scfi_sorter',
        inputHitCollection=becal_scfi_merger.outputHitCollection,
        outputHitCollection='EcalBarrelScFiHitsML',
        numberOfLayers=20,
        numberOfHits=20)

# combine layers
becal_combiner = MLDataCombiner('becal_combiner',
        inputHits1=becal_img_sorter.outputHitCollection,
        inputHits2=becal_scfi_sorter.outputHitCollection,
        outputHits='EcalBarrelCombinedHitsML',
        layerIncrement=100,
        rule='concatenate')
"""

podout.outputCommands = [
#        'keep *',
        'drop *',
        'keep EcalBarrel*Reco',
        'keep MCParticles*',
        'keep EcalBarrel*ML',
        'keep *Corr',
]

ApplicationMgr(
    TopAlg=[
        podin,
        becal_img_digi, becal_img_reco,
        # becal_img_merger, becal_img_sorter,
        becal_scfi_digi, becal_scfi_reco, becal_scfi_merger,
        # becal_scfi_sorter,
        # becal_combiner,
        podout
        ],
    EvtSel='NONE',
    EvtMax=kwargs['nev'],
    ExtSvc=[podev],
    OutputLevel=WARNING
)

