#! /usr/local/bin/python3
'''
    A python script to facilitate the ML benchmarks for e/pi separation with the imaging calorimeter (single particles).
    This process follows the steps below:
    1. Simulation to generate training samples
    2. Study and apply E/p cut to reduce the training samples
    3. Train and test ML models with the "cleaned" (after E/p cut) samples
    4. Benchmark the performance

    Author: Chao Peng (ANL)
    Date: 11/11/2022
'''
import os
import sys
import subprocess
import argparse


SDIR = os.path.dirname(os.path.realpath(__file__))
# {var} is from args
FILE_NAMES = dict(
    gen_script = os.path.join(SDIR, 'scripts', 'gen_particles.py'),
    rec_script = os.path.join(SDIR, 'options', 'imaging_ml_data.py'),
    epscan_script = os.path.join(SDIR, 'scripts', 'epcut_scan.py'),
    prep_script = os.path.join(SDIR, 'scripts', 'ml_data_preprocess.py'),
    ml_script = os.path.join(SDIR, 'scripts', 'ml_training.py'),

    sim_dir = os.path.join('{outdir}', 'sim_data'),
    epscan_dir = os.path.join('{outdir}', 'epcut'),
    ml_dir = os.path.join('{outdir}', 'ml_result'),

    gen_file = os.path.join('{outdir}', 'sim_data', '{ntag}_gen.hepmc'),
    sim_file = os.path.join('{outdir}', 'sim_data', '{ntag}_sim.edm4hep.root'),
    rec_file = os.path.join('{outdir}', 'sim_data', '{ntag}_rec.root'),
    ml_file = os.path.join('{outdir}', 'sim_data', '{ntag}_data.h5'),
)
# default values for argument parser
DEFAULT_COMPACT = os.path.join(
        os.environ.get('DETECTOR_PATH', ''),
        '{}.xml'.format(os.environ.get('DETECTOR_CONFIG', ''))
        )
# defined steps
SCRIPT_STEPS = (
    'sim',      # step 1; simulation to generate samples
    'epscan',   # step 2: e/p cut scan to find the optimized cut
    'prep',     # step 3: prepare tagging data (apply e/p cut)
    'train',    # step 4: train ML model and benchmark it
    # 'bmk',      # step 5: benchmarking
)


# simulation and reconstruction
def gen_sim_rec(**kwargs):
    # generate particles
    gen_cmd = [
        'python {gen_script} {gen_file}',
        '-n {nev}',
        '-s {seed}',
        '--angmin {angmin} --angmax {angmax}',
        '--pmin {pmin} --pmax {pmax}',
        '--particles {particles}',
        ]
    gen_cmd = ' '.join(gen_cmd).format(**kwargs).split(' ')
    subprocess.run(gen_cmd)

    # simulation
    sim_cmd = [
        'ddsim --runType batch --part.minimalKineticEnergy 1*TeV --filter.tracker edep0',
        '-v WARNING',
        '--numberOfEvents {nev}',
        # '--physics.list {physics_list}',
        '--inputFiles {gen_file}',
        '--outputFile {sim_file}',
        '--compact {compact}',
        ]
    if 'seed' in kwargs and kwargs['seed'] > 0:
        sim_cmd += ['--random.seed {seed}']
    sim_cmd = ' '.join(sim_cmd).format(**kwargs).split(' ')
    return_code = subprocess.run(sim_cmd).returncode
    print(return_code)
    if return_code is not None and return_code < 0:
        print("ERROR running simulation!")
        exit(return_code)
    subprocess.run(['rootls', '-t', kwargs['sim_file']])

    # reconstruction with juggler
    # export to environment variables (used to pass arguments to the option file)
    run_env = os.environ.copy()
    juggler_vars = [
        'JUGGLER_SIM_FILE {sim_file}',
        'JUGGLER_REC_FILE {rec_file}',
        'JUGGLER_COMPACT_PATH {compact}',
        'JUGGLER_N_EVENTS {nev}',
        ]
    lst = ' '.join(juggler_vars).format(**kwargs).split(' ')
    run_env.update({lst[i]: lst[i + 1] for i in range(0, len(lst), 2)})

    rec_cmd = 'gaudirun.py {rec_script}'.format(**kwargs).split(' ')
    print(rec_cmd)
    return_code = subprocess.run(rec_cmd, env=run_env).returncode
    print(return_code)
    if return_code is not None and return_code < 0:
        print("ERROR running juggler (reconstruction)!")
        exit(return_code)
    process = subprocess.run(['rootls', '-t', kwargs['rec_file']])



if __name__ == '__main__':

    # argument parser
    parser = argparse.ArgumentParser()

    parser.add_argument(
            '-n', '--n-events', type=int,
            dest='nev',
            default=100,
            help='number of events.'
            )
    parser.add_argument(
            '-o', '--outdir', type=str,
            dest='outdir',
            default='sim_output',
            help='output directory.'
            )
    parser.add_argument(
            '-t', '--name-tag', type=str,
            dest='ntag',
            default='imcal_ml',
            help='name tag for output files.'
            )
    parser.add_argument(
            '-c', '--compact', type=str,
            dest='compact',
            default=DEFAULT_COMPACT,
            help='path to detector compact file.'
            )
    parser.add_argument(
            '-s', '--seed', type=int,
            default=-1,
            help='random seed to child scripts (only pass it if > 0).'
            )
    parser.add_argument(
            '--batch-size', type=int,
            dest='batch',
            default=100000,
            help='batch size to process data.'
            )
    parser.add_argument(
            '--p-min', type=float,
            dest='pmin',
            default=1.5,
            help='minimum momentum of particles.'
            )
    parser.add_argument(
            '--p-max', type=float,
            dest='pmax',
            default=2.5,
            help='maximum momentum of particles.'
            )
    parser.add_argument(
            '--angle-min', type=float,
            dest='angmin',
            default=75,
            help='minimum scattering angle of particles.'
            )
    parser.add_argument(
            '--angle-max', type=float,
            dest='angmax',
            default=105,
            help='maximum scattering angle of particles.'
            )
    parser.add_argument(
            '--weight-pion', type=float,
            dest='wpi',
            default=10.,
            help='weights for pions.'
            )
    parser.add_argument(
            '--particles', type=str,
            default='electron,pion-',
            help='partcile names, separated by \",\".'
            )
    parser.add_argument(
            '--test-size', type=float,
            default=0.2,
            help='relative size of test samples (between 0-1).'
            )
    parser.add_argument(
            '--no-epcut', action='store_true',
            default=False,
            help='do not apply E/p cut to ML samples, might be useful for test runs with small samples.'
            )
    parser.add_argument(
            '--steps', type=str,
            default=', '.join(SCRIPT_STEPS),
            help='FOR DEV: choose the steps to be executed ({}).'.format(', '.join(SCRIPT_STEPS))
            )

    args = parser.parse_args()
    kwargs = vars(args)

    # prepare
    steps = [p.strip() for p in args.steps.split(',')]

    # make dirs, add paths to kwargs
    FILE_NAMES.update({key: val.format(**kwargs) for key, val in FILE_NAMES.items()})
    for key, val in FILE_NAMES.items():
        if key.endswith('_dir'):
            os.makedirs(val, exist_ok=True)
    kwargs.update(FILE_NAMES)

    # simulation for benchmark samples
    if SCRIPT_STEPS[0] in steps:
        gen_sim_rec(**kwargs)

    # ep cut scan
    if SCRIPT_STEPS[1] in steps:
        cmd = [
            'python {epscan_script} {rec_file}',
            '-o {epscan_dir} --name-tag {ntag}',
            '--batch-size {batch}',
            '-s {seed}',
            '--optimize-for 11',
            '--optimize-efficiency 0.97',
            ]
        cmd = ' '.join(cmd).format(**kwargs).split(' ')
        return_code = subprocess.run(cmd).returncode
        if return_code is not None and return_code < 0:
            print("ERROR running epcut scan script!")
            exit(return_code)

    # prepare ML data
    if SCRIPT_STEPS[2] in steps:
        cmd = [
            'python {prep_script} {rec_file}',
            '-o {ml_file}',
            '--batch-size {batch}',
            '-s {seed}',
            '--nhits 50',
            ]
        if not args.no_epcut:
            cmd.append('--epcut {epscan_dir}/{ntag}_result.json')
        cmd = ' '.join(cmd).format(**kwargs).split(' ')
        return_code = subprocess.run(cmd).returncode
        if return_code is not None and return_code < 0:
            print("ERROR running data preprocessing script!")
            exit(return_code)

    # train ML model and benchmarking
    if SCRIPT_STEPS[3] in steps:
        cmd = [
            'python {ml_script} {ml_file}',
            '-o {ml_dir} --name-tag {ntag}',
            '--weight-pion-minus {wpi}',
            '-s {seed}',
            '--batch-size 256',
            '--epochs 20',
            '--optimize-for 11',
            '--optimize-efficiency 0.98',
            '--test-size {test_size}',
            ]
        cmd = ' '.join(cmd).format(**kwargs).split(' ')
        return_code = subprocess.run(cmd).returncode
        if return_code is not None and return_code < 0:
            print("ERROR running ML training script!")
            exit(return_code)

    # TODO: collect the two-step results (e/p cut and ML)

