"""
    A script to scan the optimized cut on layer and E/p.
    It groups [event, layer] and scan the E/p cut layer-wise (e.g., EPIC brycecanyon configuration has 16 ScFi layers)
    The best 2D cut (highest rejection power) on [layer, E/p] that satisfies required effiiency will be given

    Chao Peng (ANL)
    2022/11/13
"""
import os
import gc
import sys
import json
import ROOT
import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.backends.backend_pdf as mpdf
from matplotlib.ticker import MultipleLocator
from collections import OrderedDict
from utils import flatten_collection, imcal_info, par_table, NpEncoder


# default color cycle of matplotlib
prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
            'file', type=str,
            help='path to root file.'
            )
    parser.add_argument(
            '-o', type=str,
            default='epcuts',
            dest='outdir',
            help='output directory.'
            )
    parser.add_argument(
            '--batch-size', type=int,
            default=100000,
            help='batch size to process data.'
            )
    parser.add_argument(
            '--name-tag', type=str,
            default='imcal',
            dest='ntag',
            help='nametag for output file.'
            )
    # parser.add_argument('--compact', type=str, default='', dest='compact', help='compact file')
    parser.add_argument(
            '-b', '--branch', type=str,
            default='EcalBarrelScFiGridReco',
            help='name of data branch (edm4eic::CalorimeterHitCollection).'
            )
    parser.add_argument(
            '-b2', '--truth-branch', type=str,
            default='MCParticles',
            help='name of truth mc branch.'
            )
    parser.add_argument(
            '-s', '--seed', type=int,
            default=-1,
            help='random seed.'
            )
    # flat resolution now, may need add P dependent terms?
    parser.add_argument(
            '-r', '--tracking-resolution', type=float,
            dest='res',
            default=5e-3,
            help='relative resolution of tracking (used to smear truth momentum).'
            )
    parser.add_argument(
            '--optimize-for', type=int,
            dest='opt_pid',
            default=11,
            help='optimize efficiency for the specified particle (PDGCODE).'
            )
    parser.add_argument(
            '--optimize-efficiency', type=float,
            dest='opt_eff',
            default=0.97,
            help='targeted efficiency for the cuts.'
            )
    parser.add_argument(
            '--sampling-fractions', type=str,
            default='0,0.11,2200',
            help='sampling fractions range (min,max,nbins).'
            )
    args = parser.parse_args()

    os.makedirs(args.outdir, exist_ok=True)

    if args.seed > 0:
        np.random.seed(args.seed)

    # read data and MCParticles
    rdf = ROOT.RDataFrame('events', args.file)
    # event range
    ntotal = rdf.Count().GetValue()
    ev_bins = np.append(np.arange(0, ntotal, args.batch_size), [ntotal])

    # sampling fraction bins
    samps = args.sampling_fractions.split(',')
    samp_range = (float(samps[0]), float(samps[1]))
    samp_nbins = int(samps[2])
    ep_bins = np.linspace(*samp_range, samp_nbins + 1)
    ep_centers = (ep_bins[:-1] + ep_bins[1:])/2.
    # data container
    nlayers = int(imcal_info.nlayers_scfi)
    hists = np.zeros(shape=(0, nlayers, samp_nbins))    # place holder

    # process events
    npars = pd.Series(dtype=int)
    # avoid insane amount of memory use with a large dataset
    for ev_begin, ev_end in zip(ev_bins[:-1], ev_bins[1:]):
        gc.collect()
        print('E/p cut: processing events {:d} - {:d}'.format(ev_begin, ev_end))

        dfm = flatten_collection(rdf, args.truth_branch, (int(ev_begin), int(ev_end)), cols=[
                'generatorStatus',
                'PDG',
                'momentum.x', 'momentum.y', 'momentum.z',
                'mass'
                ])

        # select incident particles
        dfm = dfm[dfm['generatorStatus'].isin([0, 1])]
        # NOTE: assumed single particles
        dfm = dfm.groupby('event').first().reset_index()
        # True momentum
        dfm.loc[:, 'P'] = np.sqrt(dfm['momentum.x']**2 + dfm['momentum.y']**2 + dfm['momentum.z']**2)
        # Smear with tracking resolution (0.5%)
        dfm.loc[:, 'Ptrack'] = dfm['P']*np.random.normal(1.0, args.res, len(dfm))

        # count how many types of particles
        npars = npars.add(dfm['PDG'].value_counts(), fill_value=0)
        sorted_pids = [p for p in par_table.keys() if p in npars.index] \
                    + [p for p in npars.index if p not in par_table.keys()]
        npars = npars[sorted_pids]
        # enlarge data container
        if len(npars) > hists.shape[0]:
            hists = np.vstack([hists, np.zeros(shape=(len(npars) - hists.shape[0],*hists.shape[1:]))])

        # reconstructed simulation hits
        df = flatten_collection(rdf, args.branch, (int(ev_begin), int(ev_end)), cols=[
                'layer',
                'energy',
                ])
        # cumulative energy sum for layer
        dfe = df.groupby(['event', 'layer'])['energy'].sum().groupby(level=0).cumsum()
        # fill missing layer (no hits in that layer) edep values with the previous value
        dfe = dfe.reset_index(level='event').groupby('event', as_index=False)\
                 .apply(lambda grp: grp.reindex(np.arange(nlayers) + 1, method='ffill'))\
                 .reset_index().drop('level_0', axis=1)
        # print(dfe)
        dfe = dfe.reset_index().merge(dfm[['event', 'Ptrack', 'PDG']], on='event')
        dfe.loc[:, 'eoverp'] = dfe['energy']/dfe['Ptrack']

        for pid, hist in zip(npars.index, hists):
            mask = dfe['PDG'] == pid
            dft = dfe[mask]
            for i in np.arange(nlayers):
                vals = dft[dft['layer'] == i + 1]['eoverp'].values
                hvals, _ = np.histogram(vals, bins=ep_bins)
                hist[i] = hist[i] + hvals

    # fill dictionary if not pre-defined
    for pid in npars.index:
        if pid not in par_table.keys():
            print('WARNING = ML training: particle ({}) not found in table, updating it.'.format(pid))
            par_table[pid] = dotdict(
                name='PDG_{:d}'.format(pid),
                full='PDG-{:d}'.format(pid),
                latex='PDG_{:d}'.format(pid)
                )

    # save binned data
    dfr = pd.DataFrame()
    for pid, hist in zip(npars.index, hists):
        dft = pd.DataFrame(
                columns=['bin_left', 'bin_right'] + ['layer_{:d}'.format(i+1) for i in np.arange(nlayers)],
                data=np.vstack([ep_bins[:-1], ep_bins[1:], hist]).T
                )
        dft.loc[:, 'PDG'] = pid
        dfr = pd.concat([dfr, dft], ignore_index=True)
    dfr.to_csv(os.path.join(args.outdir, '{}_binned_data.csv'.format(args.ntag)), index=False)

    # study E/p cut
    opt_par = par_table.get(args.opt_pid)
    if args.opt_pid not in npars.index:
        print('Error = E/p cut: cannot find targeted particle {} in samples. Stop E/p cut scan.'.format(args.opt_pid))
        exit(-1)

    # find the hist for targeted particle
    opt_idx = {p: i for i, p in enumerate(npars.index)}[args.opt_pid]
    opt_hist = hists[opt_idx]

    pdf = mpdf.PdfPages(os.path.join(args.outdir, '{}_layers.pdf'.format(args.ntag)))
    # distribution plot by layer
    box_props = dict(boxstyle='round', facecolor='white', alpha=0.5)
    cut_data = []
    for i in np.arange(nlayers):
        # cut position
        # NOTE: larger error with wider bins
        perc = np.cumsum(opt_hist[i])/np.sum(opt_hist[i])
        idx = len(perc[perc < 1. - args.opt_eff])
        ep_cut = ep_centers[idx]

        fig, ax = plt.subplots(figsize=(8, 8))
        texts = [r'Layer $\leq${:d}'.format(i + 1)]
        # cut results
        cut_result = OrderedDict(ep_cut=ep_cut)
        for k, (pid, hist) in enumerate(zip(npars.index, hists)):
            vals = hist[i]
            npar = npars[pid]
            ncut = np.sum(vals[idx:])
            eff = ncut/float(npar)
            # uncertainties: binomial variance npq
            nerr = np.sqrt(npar*eff*(1. - eff))
            # d(x/N) = dx/N
            eff_err = nerr/float(npar)

            par = par_table.get(pid)
            cut_result[par.name] = OrderedDict(
                efficiency=eff,
                error=eff_err,
                # nsamples=npar, # this is redundant
                nsamples_after_cut=ncut,
                )
            ax.hist(ep_centers, weights=vals/npar, bins=50, label='${}$'.format(par.latex),
                    color=colors[k], ec=colors[k], alpha=0.5)
            texts.append(r'$\epsilon_{{{}}}={:.2f}$%'.format(par.latex, eff*100.))
        cut_data.append(cut_result)

        ax.legend(fontsize=20, ncol=2, loc='upper center', bbox_to_anchor=(0.5, 1.12),)
        ax.tick_params(labelsize=20)
        ax.set_xlabel('$E/p$', fontsize=20)
        ax.set_ylabel('Normalized Counts', fontsize=20)
        ax.axvline(x=ep_cut, color='k', ls='--', lw=2)
        ax.text(0.5, 0.97, '\n'.join(texts),
                transform=ax.transAxes, fontsize=20, va='top', ha='center', bbox=box_props)
        pdf.savefig(fig)
        plt.close(fig)

    # 2d-scan plot
    fig, ax1 = plt.subplots(figsize=(8, 8))
    ax2 = ax1.twinx()
    ax2.set_yscale('log')

    # cut position
    cut_pos = [c.get('ep_cut') for c in cut_data]
    ax1.plot(np.arange(nlayers) + 1, cut_pos, ls='-', color=colors[0])

    # rejection power for other particles
    npars2 = npars[npars.index != args.opt_pid]
    # use alpha to identify different particles
    alphas = 1.0 - np.linspace(0, 0.7, len(npars2))

    rej_tot = np.zeros(shape=(nlayers,))
    for pid, alpha in zip(npars2.index, alphas):
        par = par_table.get(pid)
        pdata = [c.get(par.name) for c in cut_data]
        eff = np.array([c.get('efficiency') for c in pdata])
        eff_err = np.array([c.get('error') for c in pdata])
        rej_pow = np.reciprocal(eff)
        # d(1/x) = -dx/x^2
        rej_err = rej_pow**2*eff_err
        rej_tot += rej_pow * npars2[pid]/float(npars2.sum())
        # NOTE: low and high errors are switched when calculating rejection
        ax2.errorbar(np.arange(nlayers) + 1, rej_pow, yerr=rej_err,
                     fmt='o', capsize=3, color=colors[1], alpha=alpha, label='${}$'.format(par.latex))

    ax1.set_xlabel('Layer Number', fontsize=20)
    ax1.set_ylabel('Cut Position (E/p)', color=colors[0], fontsize=20)

    ax2.grid(axis='both', which='both', ls=':')
    ax2.xaxis.set_major_locator(MultipleLocator(5))
    ax2.xaxis.set_minor_locator(MultipleLocator(1))
    ax2.set_ylabel('Rejection Power', color=colors[1], fontsize=20)
    # ax2.legend(fontsize=20, ncol=4, loc='upper center', bbox_to_anchor=(0.5, 1.12),)
    ax1.set_title('2D Scan of E/p Cut', fontsize=22)
    ax1.tick_params(labelsize=20)
    ax2.tick_params(labelsize=20)
    ax1.text(0.5, 0.03, r'$\epsilon_{{{}}}\geq$ {:.2f}%'.format(opt_par.latex, args.opt_eff*100.),
             transform=ax1.transAxes, fontsize=20, va='bottom', ha='center', bbox=box_props)
    fig.subplots_adjust(left=0.15, right=0.85)
    pdf.savefig(fig)
    plt.close(fig)

    pdf.close()

    # save cut position and benchmarks
    nsamples = OrderedDict(total=npars.sum())
    for pid, count in npars.items():
        par = par_table.get(pid)
        nsamples[par.name] = count
    ep_dict = OrderedDict(
        info=OrderedDict(
            nsamples=nsamples,
            optimize_for=opt_par.name,
            optimize_efficiency=args.opt_eff,
            tracking_resolution=args.res,
            ),
        best=OrderedDict(layer=0),
        )

    for i, data in enumerate(cut_data):
        ep_dict['layer_{:d}'.format(i + 1)] = data

    # best layer
    idx = np.argmax(rej_tot)
    ep_dict['best'].update({'layer': idx + 1, 'rejection_power': rej_tot[idx]})
    ep_dict['best'].update(cut_data[idx])
    # save to json
    ep_json = json.dumps(ep_dict, indent=4, cls=NpEncoder)
    with open(os.path.join(args.outdir, '{}_result.json'.format(args.ntag)), 'w') as outfile:
        outfile.write(ep_json)

