"""
    A script to prepare datasets for ML training.
    It may also apply E/p cut to clean-up the samples

    Chao Peng (ANL)
    2022/11/13
"""
import ROOT
import os
import gc
import sys
import json
import numpy as np
import pandas as pd
import argparse
from utils import flatten_collection, cartesian_to_polar, imcal_info


pd.set_option('display.max_rows', 500)

# normalizing range for features
# NOTE: assumed for barrel ecal, using the range of 0.5 - 2 meters to normalize the r0 values
WIN_ETA = (-0.2, 0.2)   # unitless
WIN_PHI = (-0.4, 0.4)   # rad
WIN_R0 = (500, 2000)    # mm
WIN_EH = (0, 0.05)      # GeV


# only positive weights are returned
# basis defines the minimum contribution (4.6 is about 1%)
def log_weights(val, basis=4.6):
    w = np.log(val) + basis
    return np.clip(w, 0., None)


def read_epcut(path):
    if not path or not os.path.exists(path):
        print('Prepare ML data: not applying E/p cut because E/p file is not provided or missing.')
        return None
    with open(args.epcut, 'r') as f:
        data = json.load(f)
        thecut = data.get('best', {})
        thecut.update(data.get('info', {}))
        return thecut


def lin_norm(vals, window, val_name='', var_name='', warn_win_size=True):
    # statistically meaningful
    if warn_win_size and len(vals) > 1000:
        perc = (np.percentile(vals, 5), np.percentile(vals, 95))
        if perc[0] < window[0] or perc[1] > window[1]:
            warnstr = 'WARNING = Prepare ML data: normalization window {} does not fit {} values {}.'\
                      .format(window, val_name, perc)
            if var_name:
                warnstr += ' Try adjust {}'.format(var_name)
            print(warnstr)
    return np.clip((vals - window[0])/(window[1] - window[0]), 0., 1.)


# prepare ML datasets from dataframe
def format_ml_data(dfh, dft, nlayers=20, nhits=50,
                   win_eta=WIN_ETA, win_phi=WIN_PHI, win_r0=WIN_R0, win_eh=WIN_EH):
    # sanity check
    dfh = dfh[dfh['energy'] > 0.]

    # add more data
    r, theta, phi, rc, eta = cartesian_to_polar(*dfh[['position.x', 'position.y', 'position.z']].values.T)

    event_group = dfh.groupby('event')
    # convert eta, phi, edep so they are within [0, 1]
    dfh.loc[:, 'etotal'] = event_group['energy'].transform('sum')
    # get the hits energy ratio
    dfh.loc[:, 'eratio'] = dfh['energy']/dfh['etotal']

    # calculate weighted center of the event, in the future this can be replaced by tracking input
    # NOTE: assumed single particle events
    # log weighting (4.6
    dfh.loc[:, 'logw'] = log_weights(dfh['eratio'].values)
    dfh.loc[:, 'wtotal'] = event_group['logw'].transform('sum')
    # sanity check
    dfh = dfh[dfh['wtotal'] > 0.]
    # regroup because of mutation
    event_group = dfh.groupby('event')

    dfh.loc[:, 'logw'] = dfh['logw']/dfh['wtotal']
    dfh.loc[:, 'xw'] = dfh['position.x']*dfh['logw']
    dfh.loc[:, 'yw'] = dfh['position.y']*dfh['logw']
    dfh.loc[:, 'zw'] = dfh['position.z']*dfh['logw']

    # position center
    dfh.loc[:, 'xc'] = event_group['xw'].transform('sum')
    dfh.loc[:, 'yc'] = event_group['yw'].transform('sum')
    dfh.loc[:, 'zc'] = event_group['zw'].transform('sum')

    # single hits features
    r, theta, phi, r0, eta = cartesian_to_polar(*dfh[['position.x', 'position.y', 'position.z']].values.T)
    # center
    rc, thetac, phic, r0c, etac = cartesian_to_polar(*dfh[['xc', 'yc', 'zc']].values.T)

    # data for ML
    dfml = dfh[['event', 'layer']].copy()
    # NOTE: not much a difference observed between eratio or normalized E with a narrow momentum range
    #       (e.g., 1.5-2.5 GeV/c)
    dfml.loc[:, 'eh'] = dfh['eratio'].values
    # dfml.loc[:, 'eh'] = lin_norm(dfh['energy'].values, win_eh, "Ehit")
    dfml.loc[:, 'r0'] = lin_norm(r0, win_r0, "r0")
    dfml.loc[:, 'eta'] = lin_norm(eta - etac, win_eta, "eta")
    dfml.loc[:, 'phi'] = lin_norm(phi - phic, win_phi, "phi")

    # sort by energy
    dfml.sort_values(['event', 'layer', 'eh'], ascending=[True, True, False])
    dfml.loc[:, 'hit'] = dfml.groupby(['event', 'layer']).cumcount() + 1
    dfml.set_index(['event', 'layer', 'hit'], inplace=True)

    # shape data (nevents, nlayers, nhits)
    # padding with zeros
    # NOTE: if more than nhits per layer, the least energetic hits are dropped because of the previous energy sorting
    dfml = dfml.reindex(
                pd.MultiIndex.from_product(
                    [dfml.index.levels[0], np.arange(nlayers) + 1, np.arange(nhits) + 1],
                    names=['event', 'layer', 'hit']),
                fill_value=0
                )

    return dfml, list(event_group.groups.keys())


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Visualize the cluster (layer-wise) from analysis')
    parser.add_argument(
            'file', type=str,
            help='path to the input file.'
            )
    parser.add_argument(
            '-o', type=str,
            default='.',
            dest='outpath',
            help='path to output data file (hdf5).'
            )
    parser.add_argument(
            '--batch-size', type=int,
            default=100000,
            help='batch size to process data.'
            )
    parser.add_argument(
            '-s', '--seed', type=int,
            default=-1,
            help='random seed.'
            )
    parser.add_argument(
            '--nhits', type=int,
            default=50,
            dest='nhits',
            help='number of hits per layer.'
            )
    parser.add_argument(
            '-b', '--img-branch', type=str,
            default='EcalBarrelImagingHitsReco',
            help='name of data branch (edm4eic::CalorimeterHitCollection).'
            )
    parser.add_argument(
            '-b2', '--scfi-branch', type=str,
            default='EcalBarrelScFiGridReco',
            help='name of data branch (edm4eic::CalorimeterHitCollection).'
            )
    parser.add_argument(
            '-b3', '--truth-branch', type=str,
            default='MCParticles',
            help='name of truth mc branch.'
            )
    parser.add_argument(
            '--epcut', type=str,
            default=None,
            help='A json file contains epcut information (best will be used).'
            )
    parser.add_argument(
            '--append-data', action='store_true',
            default=False,
            help='Append data to pre-existing data file.'
            )
    args = parser.parse_args()

    data_store = pd.HDFStore(args.outpath)
    # print(data_store.keys())
    # clear data store
    if not args.append_data:
        for dkey in [imcal_info.ml_data_key, imcal_info.truth_data_key]:
            if dkey in data_store.keys():
                data_store.remove(dkey)
                print('Prepare ML data: remove pre-existed dataset {}'.format(dkey))

    if args.seed > 0:
        np.random.seed(args.seed)

    # read data and MCParticles
    rdf = ROOT.RDataFrame("events", args.file)
    # event range
    ntotal = rdf.Count().GetValue()
    ev_bins = np.append(np.arange(0, ntotal, args.batch_size), [ntotal])

    # epcut info
    epcut = read_epcut(args.epcut)
    # print(epcut)

    # process events
    # avoid insane amount of memory use with a large dataset
    for ev_begin, ev_end in zip(ev_bins[:-1], ev_bins[1:]):
        gc.collect()
        print('Prepare ML data: processing events {:d} - {:d}'.format(ev_begin, ev_end))

        dfm = flatten_collection(rdf, args.truth_branch, (int(ev_begin), int(ev_end)), cols=[
                'generatorStatus',
                'PDG',
                'momentum.x', 'momentum.y', 'momentum.z',
                'mass'
                ])

        # select incident particles
        dfm = dfm[dfm['generatorStatus'].isin([0, 1])]
        # NOTE: assumed single particles
        dfm = dfm.groupby('event').first()
        # True momentum
        dfm.loc[:, 'P'] = np.sqrt(dfm['momentum.x']**2 + dfm['momentum.y']**2 + dfm['momentum.z']**2)
        # print(dfm)

        # scfi information
        dfs = flatten_collection(rdf, args.scfi_branch, (int(ev_begin), int(ev_end)), cols=[
                'layer',
                'energy',
                'position.x', 'position.y', 'position.z',
                ])

        #  information
        dfi = flatten_collection(rdf, args.img_branch, (int(ev_begin), int(ev_end)), cols=[
                'layer',
                'energy',
                'position.x', 'position.y', 'position.z',
                ])

        # apply ep_cut
        if epcut:
            # cumulative energy sum for layer
            dfe = dfs.groupby(['event', 'layer'])['energy'].sum().groupby(level=0).cumsum()
            # fill missing layer (no hits in that layer) edep values with the previous value
            dfe = dfe.reset_index(level='event').groupby('event', as_index=False)\
                     .apply(lambda grp: grp.reindex(np.arange(imcal_info.nlayers_scfi) + 1, method='ffill'))\
                     .reset_index().drop('level_0', axis=1)
            # unstack layer edeps
            dfe = dfe.set_index(['event', 'layer'])\
                     .stack().unstack(level='layer')
            dfe.columns = ['layer_{}'.format(i) for i in dfe.columns]
            dfe.columns.name = None
            dfe = dfe.reset_index().drop('level_1', axis=1).set_index('event', drop=True)
            # print(dfe)

            p_smear = dfm['P']*np.random.normal(1.0, epcut['tracking_resolution'], len(dfm))
            ep_mask = (dfe['layer_{layer}'.format(**epcut)]/p_smear) >= epcut['ep_cut']
            # print(np.sum(ep_mask))
            dfm = dfm[ep_mask]
            dfs = dfs[dfs['event'].isin(dfm.index)]
            dfi = dfi[dfi['event'].isin(dfm.index)]

        # reformat data for ML
        dfi, ievs = format_ml_data(dfi, dfm, nlayers=imcal_info.nlayers_img, nhits=args.nhits)
        dfs, sevs = format_ml_data(dfs, dfm, nlayers=imcal_info.nlayers_scfi, nhits=args.nhits)
        if ievs != sevs:
            common = np.intersect1d(ievs, sevs)
            dfi = dfi[dfi.index.isin(common, level='event')]
            dfs = dfs[dfs.index.isin(common, level='event')]
            dfm = dfm[dfm.index.isin(common)].sort_index()
        else:
            dfm = dfm[dfm.index.isin(ievs)].sort_index()

        # assign layer type to different dataset
        dfi.loc[:, 'ltype'] = 'img'
        dfi.loc[:, 'lval'] = 0
        dfs.loc[:, 'ltype'] = 'scfi'
        dfs.loc[:, 'lval'] = 1

        # null eta value for ScFi layer
        # NOTE: add it back in the future with timing-z resolution
        dfs['eta'] = 0.

        # combine two datasets
        df = pd.concat([dfi.reset_index(), dfs.reset_index()], ignore_index=True)\
               .set_index(['event', 'ltype', 'layer', 'hit']).sort_index()
        # print(df.head(100))
        data_store.append(imcal_info.ml_data_key, df, format='t',  data_columns=True)
        data_store.append(imcal_info.truth_data_key, dfm[['PDG', 'P', 'mass']], format='t', data_columns=True)

    data_store.close()


