"""
    Script for supervised training of a simple ML model using Tensorflow 2.
    ML e/pi separation for imaging barrel ecal.

    Chao Peng (ANL)
    2022/11/11
"""
import os
import sys
import json
import argparse
import pandas as pd
import numpy as np
from collections import OrderedDict

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FixedLocator, MaxNLocator
from matplotlib.colors import LogNorm

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers

from utils import imcal_info, par_table, NpEncoder


# default color cycle of matplotlib
prop_cycle = plt.rcParams['axes.prop_cycle']
colors = prop_cycle.by_key()['color']


def count_pid_labels(label_lst, pid_list):
    res = OrderedDict()
    res['total'] = len(label_lst)
    for p, cnt in zip(pid_list, np.bincount(label_lst)):
        res[par_table.get(p).name] = '{:d}'.format(cnt)
    return res


def build_model():
    my_model = keras.Sequential([
        keras.layers.Conv2D(64, (3, 3), padding='same', activation='relu', input_shape=xdata.shape[1:]),
        keras.layers.MaxPooling2D((2, 2), strides=2),
        keras.layers.Dropout(0.25),
        keras.layers.Conv2D(128, (2, 2), padding='same', activation='relu'),
        keras.layers.MaxPooling2D((2, 2), strides=2),
        keras.layers.Conv2D(64, (2, 2), padding='same', activation='relu'),
        keras.layers.MaxPooling2D((2, 2), strides=2),
        keras.layers.Dropout(0.25),

        keras.layers.Flatten(),
        keras.layers.Dense(128, activation='relu'),
        # keras.layers.Dropout(0.25),
        keras.layers.Dense(32, activation='relu'),
        keras.layers.Dense(len(pid_labels), activation='softmax')
        ])
    return my_model


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
            'data_store', type=str,
            help='path to data store.'
            )
    parser.add_argument(
            '-o', type=str,
            dest='outdir',
            default='ml_result',
            help='output directory.'
            )
    parser.add_argument(
            '-s', '--seed', type=int,
            default=-1,
            help='random seed.'
            )
    parser.add_argument(
            '-t', '--name-tag', type=str,
            dest='ntag',
            default='imcal_ml',
            help='name tag for output files/models.'
            )
    parser.add_argument(
            '--test-size', type=float,
            default=0.2,
            help='relative size of test samples (between 0-1).'
            )
    parser.add_argument(
            '--epochs', type=int,
            default=20,
            help='epochs for ML training.'
            )
    parser.add_argument(
            '--batch-size', type=int,
            default=512,
            help='batch size for ML training.'
            )
    parser.add_argument(
            '--optimize-for', type=int,
            default=11,
            help='optimize cut for a target particle (PDGCODE).'
            )
    parser.add_argument(
            '--optimize-efficiency', type=float,
            default=0.98,
            help='used when --optimize-for is specified, try to keep the efficiency with optimized cuts.'
            )
    parser.add_argument(
            '--read-model', action='store_true',
            default=False,
            help='read a trained model'
            )
    parser.add_argument(
            '--read-model-path', type=str,
            default='',
            help='only used when --read-model is enabled, if not specified it reads the model from saving path.'
            )
    # weights
    for key, val in par_table.items():
        parser.add_argument(
            '--weight-{}'.format(val.full), type=float,
            default=1.0,
            help='sample weight for {}.'.format(val.name)
            )
    args = parser.parse_args()
    kwargs = vars(args)

    os.makedirs(args.outdir, exist_ok=True)

    if args.seed > 0:
        np.random.seed(args.seed)
        tf.random.set_seed(args.seed)

    df = pd.read_hdf(args.data_store, key=imcal_info.ml_data_key)
    # NOTE: assumed event index is exactly the same as df
    dft = pd.read_hdf(args.data_store, key=imcal_info.truth_data_key)
    print('ML training: input sample size.')
    print(dft['PDG'].value_counts())

    # fill dictionary if not pre-defined
    data_pid = dft['PDG'].unique()
    for pid in data_pid:
        if pid not in par_table.keys():
            print('WARNING = ML training: particle ({}) not found in table, updating it.'.format(pid))
            par_table[pid] = dotdict(
                name='PDG_{:d}'.format(pid),
                full='PDG-{:d}'.format(pid),
                latex='PDG_{:d}'.format(pid)
                )

    # build labels and weights for particles
    # NOTE: pid follows the dictionary orders
    pid_labels = OrderedDict()
    pid_weights = OrderedDict()
    for pid, part in par_table.items():
        if pid not in data_pid:
            continue
        pid_weights[pid] = kwargs.get('weight_{}'.format(part.full.replace('-', '_')), 1.0)
        pid_labels[pid] = len(pid_labels)
    pids = list(pid_labels.keys())
    # print(pid_labels)
    # print(pid_weights)

    # data shape
    # NOTE: number of layers for combined data
    nlayers = imcal_info.nlayers_img + imcal_info.nlayers_scfi
    # NOTE: index levels should be [event, layer_type, layer, hit], change the code if data structure is changed
    nhits = df.index.levels[-1].max()

    # training datasets
    xdata = df.values.reshape(dft.shape[0], nlayers, nhits, len(df.columns))
    ydata = dft['PDG'].map(pid_labels).values
    wdata = dft['PDG'].map(pid_weights).values

    # seperate train and test data
    indices = np.arange(len(xdata))
    np.random.shuffle(indices)
    id_train = indices[int(len(indices)*args.test_size):]
    id_test = indices[:int(len(indices)*args.test_size)]
    # test data will be used for benchmarking, no need for sampling weights
    xtrain, ytrain, wtrain = xdata[id_train], ydata[id_train], wdata[id_train]
    xtest, ytest = xdata[id_test], ydata[id_test]

    # try to use GPUs
    # tf.debugging.set_log_device_placement(True)
    gpus = tf.config.list_logical_devices('GPU')
    print("ML training: number of GPUs available: ", len(gpus))
    strategy = tf.distribute.MirroredStrategy(gpus)
    if args.read_model:
        model_path = os.path.join(args.outdir, '{}_model'.format(args.ntag))
        if args.read_model_path:
            model_path = args.read_model_path
        model = keras.models.load_model(model_path)
    else:
        with strategy.scope():
            train_dataset = tf.data.Dataset.from_tensor_slices((xtrain, ytrain, wtrain))
            model = build_model()
            model.compile(optimizer=keras.optimizers.Adam(learning_rate=1e-4),
                          loss=keras.losses.SparseCategoricalCrossentropy(from_logits=False),
                          metrics=['accuracy'])
            history = model.fit(train_dataset.batch(args.batch_size), epochs=args.epochs)
            model.save(os.path.join(args.outdir, '{}_model'.format(args.ntag)))

    # benchmark
    prediction = model.predict(xtest)
    pred_weights = np.ones(len(pid_labels))

    # results
    npar = len(pid_labels)
    opt_i = 0
    opt_cut = 0.

    # optimize for target particle's efficiency
    opt_param = (None, 0, 1.0)
    if args.optimize_for != 0:
        opt_pid = args.optimize_for
        if opt_pid not in pid_labels.keys():
            print('WARNING = ML training: ' +
                  'particle ({}) [PDGCODE] not found in samples, skip optimizing.'.format(opt_pid))
        else:
            opt_i = pid_labels.get(opt_pid)
            # find optimal prob weight
            opt_probs = prediction[ytest == opt_i].T[opt_i]
            opt_cut = np.percentile(opt_probs, (1. - args.optimize_efficiency)*100.)
            # NOTE: this weight calculation only works for two labels
            opt_wt = (1. - opt_cut)/opt_cut
            pred_weights[opt_i] = opt_wt
            opt_param = (par_table.get(opt_pid).name, args.optimize_efficiency, opt_wt)

    # pick labels with the highest probability
    pred_labels = np.argmax(prediction*pred_weights, axis=1)
    pred_probs = np.zeros(shape=(npar, npar))
    opt_label = par_table.get(pids[opt_i]).latex

    fig, ax = plt.subplots(figsize=(12, 9), dpi=160)
    effs = []
    for pid, ilbl in pid_labels.items():
        mask = (ytest == ilbl)
        pred_probs[ilbl] = np.bincount(pred_labels[mask])/float(np.sum(mask))
        labels = par_table.get(pid)

        ax.hist(prediction[mask].T[opt_i], bins=np.linspace(0, 1, 101), label='${}$'.format(labels.latex),
                color=colors[ilbl], ec=colors[ilbl], alpha=0.5)

        effs.append((labels, pred_probs[ilbl][opt_i]))
    if opt_cut > 0.:
        ax.axvline(x=opt_cut, lw=2, color='k', ls='--')
        eff_string = '\n'.join([r'$\epsilon_{{{}}} = {:.2f}$%'.format(labels.latex, eff*100.) for labels, eff in effs])
        data_to_axis = (ax.transAxes + ax.transData.inverted()).inverted()
        ax.text(data_to_axis.transform((opt_cut, 1))[0] + 0.01, 0.99, eff_string, fontsize=24,
                transform=ax.transAxes, ha='left', va='top')
    ax.set_yscale('log')
    ax.set_ylabel('Counts', fontsize=24)
    ax.set_xlabel(r'$P_{{{}}}$'.format(opt_label), fontsize=24)
    ax.tick_params(direction='in', which='both', labelsize=24)
    ax.legend(fontsize=24, ncol=4, loc='upper center', bbox_to_anchor=(0.5, 1.12),)
    fig.savefig(os.path.join(args.outdir, '{}_pid_hist.pdf'.format(args.ntag)))

    # save results
    # all results
    result = OrderedDict(
            training= OrderedDict(
                epochs=args.epochs,
                batch_size=args.batch_size,
                sample_size=count_pid_labels(ytrain, pids),
                sample_weights={par_table.get(pid).name: weight for pid, weight in pid_weights.items()},
                optimal_cut=OrderedDict(
                    target_particle=opt_param[0],
                    target_efficiency=opt_param[1],
                    target_label_weight=opt_param[2],
                    ),
                ),
            test=OrderedDict(
                sample_size=count_pid_labels(ytest, pids),
                efficiencies=OrderedDict([(labels.name, eff) for labels, eff in effs]),
                rejections=OrderedDict([(labels.name, 1./eff) for labels, eff in effs]),
                ),
            )
    res_json = json.dumps(result, indent=4, cls=NpEncoder)
    with open(os.path.join(args.outdir, '{}_result.json'.format(args.ntag)), 'w') as outfile:
        outfile.write(res_json)

