'''
    functions and objects that are commonly used by all scripts within this benchmark

    Chao Peng (ANL)
    2022/11/13
'''
import json
import pandas as pd
import numpy as np
from collections import OrderedDict


class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


# basic information share among all scripts
imcal_info = dotdict(
        nlayers_img=6,
        nlayers_scfi=16,
        ml_data_key='/ml_data',
        truth_data_key='/truth',
    )

# a lookup table to convert pdgcode to text
par_table = OrderedDict()
par_table[11] = dotdict(name='electron', full='electron', latex='e^-')
par_table[-211] = dotdict(name='pion-', full='pion-minus', latex='\pi^-')


# an encoder to serialize numpy data
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


# read from RDataFrame and flatten a given collection, return pandas dataframe
def flatten_collection(rdf, collection, ev_range=None, cols=None, event_colname='event'):
    if not cols:
        cols = [str(c) for c in rdf.GetColumnNames() if str(c).startswith('{}.'.format(collection))]
    else:
        cols = ['{}.{}'.format(collection, c) for c in cols]
    if not cols:
        print('cannot find any branch under collection {}'.format(collection))
        return pd.DataFrame()
    # take data
    if not ev_range:
        data = rdf.AsNumpy(cols)
    else:
        data = rdf.Range(*ev_range).AsNumpy(cols)
    # flatten the data, add an event id to identify clusters from different events
    evns = []
    for i, vec in enumerate(data[cols[0]]):
        evns += [i]*vec.size()
    for n, vals in data.items():
        # make sure ints are not converted to floats
        typename = vals[0].__class__.__name__.lower()
        dtype = np.int64 if 'int' in typename or 'long' in typename else np.float32
        # type safe creation
        data[n] = np.asarray([v for vec in vals for v in vec], dtype=dtype)
    # build data frame
    dfp = pd.DataFrame({c: pd.Series(v) for c, v in data.items()})
    evns = np.asarray(evns)
    if ev_range is not None:
        evns += ev_range[0]
    dfp.loc[:, event_colname] = evns
    dfp.rename(columns={c: c.replace(collection + '.', '') for c in dfp.columns}, inplace=True)
    return dfp


def cartesian_to_polar(x, y, z):
    r = np.sqrt(x**2 + y**2 + z**2)
    rc = np.sqrt(x**2 + y**2)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)
    eta = -np.log(np.tan(theta / 2.))
    return r, theta, phi, rc, eta


