'''
    A simple analysis script to extract some basic info of Monte-Carlo particles and clusters
'''
import os
import ROOT
import pandas as pd
import numpy as np
import argparse
from matplotlib import pyplot as plt
import matplotlib.ticker as ticker


# load root macros, input is an argument string
def load_root_macros(arg_macros):
    for path in arg_macros.split(','):
        path = path.strip()
        if os.path.exists(path):
            gROOT.Macro(path)
            print('Loading root macro: \'{}\' loaded.'.format(path))
        else:
            print('Loading root macros: \'{}\' does not exist, skip loading it.'.format(path))


# read from RDataFrame and flatten a given collection, return pandas dataframe
def flatten_collection(rdf, collection, cols=None):
    if not cols:
        cols = [str(c) for c in rdf.GetColumnNames() if str(c).startswith('{}.'.format(collection))]
    else:
        cols = ['{}.{}'.format(collection, c) for c in cols]
    if not cols:
        print('cannot find any branch under collection {}'.format(collection))
        return pd.DataFrame()

    data = rdf.AsNumpy(cols)
    # flatten the data, add an event id to identify clusters from different events
    evns = []
    for i, vec in enumerate(data[cols[0]]):
        evns += [i]*vec.size()
    for n, vals in data.items():
        # make sure ints are not converted to floats
        typename = vals[0].__class__.__name__.lower()
        dtype = np.int64 if 'int' in typename or 'long' in typename else np.float64
        # type safe creation
        data[n] = np.asarray([v for vec in vals for v in vec], dtype=dtype)
    # build data frame
    dfp = pd.DataFrame({c: pd.Series(v) for c, v in data.items()})
    dfp.loc[:, 'event'] = evns
    return dfp

def thrown_particles_figure(rdf, save, mcbranch="MCParticles"):
    # define truth particle info
    dft = flatten_collection(rdf, mcbranch, ['generatorStatus', 'PDG', 'momentum.x', 'momentum.y', 'momentum.z', 'mass'])
    dft.rename(columns={c: c.replace(mcbranch + '.', '') for c in dft.columns}, inplace=True)
    # select thrown particles
    dft = dft[dft['generatorStatus'] == 1]

    # figure
    fig, axs = plt.subplots(2, 2, figsize=(16, 12), dpi=120)
    # convert pid to particle names
    pdgbase = ROOT.TDatabasePDG()
    get_pname = np.vectorize(lambda pid: pdgbase.GetParticle(int(pid)).GetName())
    get_pcharge = np.vectorize(lambda pid: pdgbase.GetParticle(int(pid)).Charge()/3.)

    # enumerate particle names
    dft.loc[:, 'pname'] = get_pname(dft['PDG'].values)
    dft.loc[:, 'charge'] = get_pcharge(dft['PDG'].values)
    penum = {pname: i for i, pname in enumerate(dft['pname'].unique())}
    dft.loc[:, 'pname_id'] = dft['pname'].map(penum)

    ax = axs.flat[0]
    ax.hist(dft['pname_id'].values, bins=np.arange(-4, len(penum)*2 + 4), align='left', ec='k')
    ax.set_ylabel('Particle Counts', fontsize=24)
    ax.tick_params(labelsize=22)
    ax.set_axisbelow(True)
    ax.grid(linestyle=':', which='both', axis='y')
    ax.xaxis.set_major_locator(ticker.FixedLocator(list(penum.values())))
    ax.set_xticklabels(list(penum), rotation=30)

    # calculate kinematics
    get_4vecs = np.vectorize(lambda x, y, z, m: ROOT.Math.PxPyPzMVector(x, y, z, m))
    fourvecs = get_4vecs(*dft[['momentum.x', 'momentum.y', 'momentum.z', 'mass']].values.T)

    dft.loc[:, 'p'] = [v.P() for v in fourvecs]
    dft.loc[:, 'theta'] = [v.Theta() for v in fourvecs]
    dft.loc[:, 'phi'] = [v.Phi() for v in fourvecs]
    dft.loc[:, 'eta'] = [v.Eta() for v in fourvecs]
    dft.loc[:, 'pt'] = [v.Pt() for v in fourvecs]

    # column name, x label, y label, bins
    plots = [
        ('p', r'$P$ (GeV)', 'Counts', 50),
        ('eta', r'$\eta$', 'Counts', 50),
        ('pt', r'$P_T$ (GeV)', 'Counts', 50),
    ]
    for ax, (col, xlabel, ylabel, bins) in zip(axs.flat[1:], plots):
        ax.hist(dft[col].values, bins=bins, align='left', ec='k')
        ax.tick_params(labelsize=22, direction='in', which='both')
        ax.grid(linestyle=':', which='both')
        ax.set_axisbelow(True)
        ax.set_xlabel(xlabel, fontsize=24)
        ax.set_ylabel(ylabel, fontsize=24)

    fig.text(0.5, 0.95, 'Thrown Particles', ha='center', fontsize=24)
    fig.savefig(save)
    plt.close(fig)
    return dft

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('rec_file', help='Path to reconstruction output file.')
    parser.add_argument('-o', dest='outdir', default='.', help='Output directory.')
    parser.add_argument('-t', '--nametag', type=str, default='tracking', help='Name tag for output files.')
    parser.add_argument('-m', '--macros', dest='macros', default='rootlogon.C',
            help='Macro files to be loaded by root, separated by \",\".')
    parser.add_argument('--mc-collection', dest='mc', default='MCParticles',
            help='Collection name of MC particles truth info.')
    parser.add_argument('--tracking-collection', dest='coll', default='outputTrackParameters',
            help='Collection name of clusters to plot')
    args = parser.parse_args()

    # multi-threading for RDataFrame
    nthreads = os.cpu_count()//2
    ROOT.ROOT.EnableImplicitMT(nthreads)
    print('CPU available {}, using {}'.format(os.cpu_count(), nthreads))

    # declare some functions from c++ script, needed for data frame processing
    # script_dir = os.path.dirname(os.path.realpath(__file__))
    # fcode = open(os.path.join(script_dir, 'barrel_clusters.cxx')).read()
    # ROOT.gInterpreter.Declare(fcode)

    os.makedirs(args.outdir, exist_ok=True)
    # load macros (add libraries/headers root cannot automatically found here)
    load_root_macros(args.macros)

    rdf_rec = ROOT.RDataFrame('events', args.rec_file)
    dfm = thrown_particles_figure(rdf_rec, save=os.path.join(args.outdir, '{}_thrown_particles.png'.format(args.nametag)), mcbranch=args.mc)
    df = flatten_collection(rdf_rec, args.coll)
    df.rename(columns={c: c.replace(args.coll + '.', '') for c in df.columns}, inplace=True)

    fig, axs = plt.subplots(3, 2, figsize=(16, 18), dpi=120)
    for ax in axs.flat:
        ax.tick_params(direction='in', which='both', labelsize=20)
        ax.grid(linestyle=':')
        ax.set_axisbelow(True)
    ax = axs.flat[0]
    # tracking efficiency
    eta_bins = np.linspace(-4, 4, 21)
    sim_eta, _ = np.histogram(dfm['eta'].values, bins=eta_bins)
    rec_eta, _ = np.histogram(-np.log(np.tan(df['theta'].values/2.)), bins=eta_bins)
    eta_centers = (eta_bins[1:] + eta_bins[:-1])/2.
    eta_binsize = np.mean(np.diff(eta_centers))
    # cut off not simulated bins
    mask = sim_eta > 0
    sim_eta = sim_eta[mask]
    rec_eta = rec_eta[mask]
    eta_centers = eta_centers[mask]
    track_eff = rec_eta/sim_eta
    # binary distribution, pq*sqrt(N)
    # TODO check the errors
    eff = np.mean(track_eff)
    track_err = eff*(1. - eff)*np.reciprocal(np.sqrt(sim_eta))
    # rec_err = eff*(1. - eff)*np.sqrt(rec_eta)
    # track_eff_lower = track_eff - np.maximum(np.zeros(shape=rec_eta.shape), (rec_eta - rec_err)/sim_eta)
    # track_eff_upper = np.minimum(np.ones(shape=rec_eta.shape), (rec_eta + rec_err)/sim_eta) - track_eff
    track_eff_lower = track_eff - np.maximum(np.zeros(shape=rec_eta.shape), track_eff - track_err)
    track_eff_upper = np.minimum(np.ones(shape=rec_eta.shape), track_eff + track_err) - track_eff


    ax.errorbar(eta_centers, track_eff, xerr=eta_binsize/2., yerr=[np.abs(track_eff_lower), np.abs(track_eff_upper)],
                fmt='o', capsize=3)
    ax.set_ylim(0., 1.1)
    ax.set_xlim(-4.5, 4.5)
    ax.set_ylabel('Tracking Efficiency', fontsize=20)
    ax.set_xlabel('$\eta$', fontsize=20)

    # momentum resolution
    ax = axs.flat[1]
    dfc = dfm.set_index('event').loc[df['event'].values]
    rec_p = dfc['charge'].values/df['qOverP'].values
    sim_p = dfc['p'].values
    # print(dfc['charge'].values, df['qOverP'].values)
    dp_p = (rec_p - sim_p)/sim_p
    hval, hbins, _ = ax.hist(dp_p*100., bins=np.linspace(-20, 20, 101),
                             weights=np.repeat(1./float(rec_p.shape[0]), rec_p.shape[0]), ec='k')
    nbins = hbins.shape[0] - 1
    ax.set_ylabel('Normalized Counts / {:d} Bins'.format(nbins), fontsize=20)
    ax.set_xlabel(r'$\delta$ p (%)', fontsize=20)

    # theta resolution
    ax = axs.flat[2]
    rec_th = df.groupby('event')['theta'].first().values
    sim_th = dfc.groupby('event')['theta'].first().values
    dth_th = (rec_th - sim_th)
    # print(np.vstack([rec_th, sim_th]).T)
    hval, hbins, _ = ax.hist(dth_th, bins=np.linspace(-0.01, 0.01, 101),
                             weights=np.repeat(1./float(rec_th.shape[0]), rec_th.shape[0]), ec='k')
    nbins = hbins.shape[0] - 1
    ax.set_ylabel('Normalized Counts / {:d} Bins'.format(nbins), fontsize=20)
    ax.set_xlabel(r'$d\theta$ (rad)', fontsize=20)

    # phi resolution
    ax = axs.flat[3]
    rec_th = df.groupby('event')['phi'].first().values
    sim_th = dfc.groupby('event')['phi'].first().values
    dth_th = (rec_th - sim_th)
    # print(np.vstack([rec_th, sim_th]).T)
    hval, hbins, _ = ax.hist(dth_th, bins=np.linspace(-0.01, 0.01, 101),
                             weights=np.repeat(1./float(rec_th.shape[0]), rec_th.shape[0]), ec='k')
    nbins = hbins.shape[0] - 1
    ax.set_ylabel('Normalized Counts / {:d} Bins'.format(nbins), fontsize=20)
    ax.set_xlabel(r'$d\phi$ (rad)', fontsize=20)

    # eta distribution
    ax = axs.flat[4]
    sim_eta =  dfm.groupby('event')['eta'].first().values
    rec_eta =  -np.log(np.tan(df.groupby('event')['theta'].first().values/2.))
    hval, hbins, _ = ax.hist(sim_eta, bins=np.linspace(-4, 4, 41), ec='k',alpha=0.3,label="Generated")
    hval, hbins, _ = ax.hist(rec_eta, bins=np.linspace(-4, 4, 41), ec='k',alpha=0.3, label="Reconstructed")
    nbins = hbins.shape[0] - 1
    ax.set_ylabel('Normalized Counts / {:d} Bins'.format(nbins), fontsize=20)
    ax.set_xlabel(r'$\eta$', fontsize=20)
    ax.legend(loc="upper right")
    
    # theta distribution
    ax = axs.flat[5]
    sim_th_deg =  dfm.groupby('event')['theta'].first().values*180/np.pi
    rec_th_deg = df.groupby('event')['theta'].first().values*180/np.pi

    hval, hbins, _ = ax.hist(sim_th_deg, bins=np.linspace(-0, 180,61),
                              ec='k',alpha=0.3,label="Generated")
    hval, hbins, _ = ax.hist(rec_th_deg, bins=np.linspace(-0,180, 61),
                              ec='k',alpha=0.3, label="Reconstructed")
    nbins = hbins.shape[0] - 1
    ax.set_ylabel('Normalized Counts / {:d} Bins'.format(nbins), fontsize=20)
    ax.set_xlabel(r'$\theta (degree)$', fontsize=20)


    fig.text(0.5, 0.95, 'Tracker Benchmark (Truth Init.)', fontsize=22, ha='center')
    fig.savefig(os.path.join(args.outdir, '{}_performance.png'.format(args.nametag)))


